#!/usr/bin/env python

from neml.math import tensors, rotations
from neml.cp import crystallography

from common import differentiate

import unittest
import numpy as np
import numpy.linalg as la

ktw_tetra = np.array(
    [
    [ [ 1, 0, 0],
      [ 0, 1, 0],
      [ 0, 0, 1] ],
    [ [-1, 0, 0],
      [ 0, 1, 0],
      [ 0, 0,-1] ],
    [ [ 1, 0, 0],
      [ 0,-1, 0],
      [ 0, 0,-1] ],
    [ [-1, 0, 0],
      [ 0,-1, 0],
      [ 0, 0, 1] ],
    [ [ 0, 1, 0],
      [-1, 0, 0],
      [ 0, 0, 1] ],
    [ [ 0,-1, 0],
      [ 1, 0, 0],
      [ 0, 0, 1] ],
    [ [ 0, 1, 0],
      [ 1, 0, 0],
      [ 0, 0,-1] ],
    [ [ 0,-1, 0],
      [-1, 0, 0],
      [ 0, 0,-1] ]
    ], dtype = float)

a = np.sqrt(3.0)/2.0
h = 1.0/2.0
ktw_hexagonal = np.array(
    [
      [ [ 1, 0, 0], 
        [ 0, 1, 0],
        [ 0, 0, 1] ],
      [ [-h, a, 0],
        [-a,-h, 0],
        [ 0, 0, 1] ],
      [ [-h,-a, 0],
        [ a,-h, 0],
        [ 0, 0, 1] ],
      [ [ h, a, 0],
        [-a, h, 0],
        [ 0, 0, 1] ],
      [ [-1, 0, 0],
        [ 0,-1, 0],
        [ 0, 0, 1] ],
      [ [ h,-a, 0],
        [ a, h, 0],
        [ 0, 0, 1] ],
      [ [-h,-a, 0],
        [-a, h, 0],
        [ 0, 0,-1] ],
      [ [ 1, 0, 0],
        [ 0,-1, 0],
        [ 0, 0,-1] ],
      [ [-h, a, 0],
        [ a, h, 0],
        [ 0, 0,-1] ],
      [ [ h, a, 0],
        [ a,-h, 0],
        [ 0, 0,-1] ],
      [ [-1, 0, 0],
        [ 0, 1, 0],
        [ 0, 0,-1] ],
      [ [ h,-a, 0],
        [-a,-h, 0],
        [ 0, 0,-1 ] ]
      ])

ktw_cubic = np.array(
    [
      [ [ 1, 0, 0],
        [ 0, 1, 0],
        [ 0, 0, 1] ],
      [ [ 0, 0, 1],
        [ 1, 0, 0],
        [ 0, 1, 0] ],
      [ [ 0, 1, 0],
        [ 0, 0, 1],
        [ 1, 0, 0] ],
      [ [ 0,-1, 0],
        [ 0, 0, 1],
        [-1, 0, 0] ],
      [ [ 0,-1, 0],
        [ 0, 0,-1],
        [ 1, 0, 0] ],
      [ [ 0, 1, 0],
        [ 0, 0,-1],
        [-1, 0, 0] ],
      [ [ 0, 0,-1],
        [ 1, 0, 0],
        [ 0,-1, 0] ],
      [ [ 0, 0,-1],
        [-1, 0, 0],
        [ 0, 1, 0] ],
      [ [ 0, 0, 1],
        [-1, 0, 0],
        [ 0,-1, 0] ],
      [ [-1, 0, 0],
        [ 0, 1, 0],
        [ 0, 0,-1] ],
      [ [-1, 0, 0],
        [ 0,-1, 0],
        [ 0, 0, 1] ],
      [ [ 1, 0, 0],
        [ 0,-1, 0],
        [ 0, 0,-1] ],
      [ [ 0, 0,-1],
        [ 0,-1, 0],
        [-1, 0, 0] ],
      [ [ 0, 0, 1],
        [ 0,-1, 0],
        [ 1, 0, 0] ],
      [ [ 0, 0, 1],
        [ 0, 1, 0],
        [-1, 0, 0] ],
      [ [ 0, 0,-1],
        [ 0, 1, 0],
        [ 1, 0, 0] ],
      [ [-1, 0, 0],
        [ 0, 0,-1],
        [ 0,-1, 0] ],
      [ [ 1, 0, 0],
        [ 0, 0,-1],
        [ 0, 1, 0] ],
      [ [ 1, 0, 0],
        [ 0, 0, 1],
        [ 0,-1, 0] ],
      [ [-1, 0, 0],
        [ 0, 0, 1],
        [ 0, 1, 0] ],
      [ [ 0,-1, 0],
        [-1, 0, 0],
        [ 0, 0,-1] ],
      [ [ 0, 1, 0],
        [-1, 0, 0],
        [ 0, 0, 1] ],
      [ [ 0, 1, 0],
        [ 1, 0, 0],
        [ 0, 0,-1] ],
      [ [ 0,-1, 0],
        [ 1, 0, 0],
        [ 0, 0, 1] ]
      ])

def ktw_rotational_groups(typ):
  if typ == "1":
    return ktw_tetra[0:1]
  elif typ == "2":
    return ktw_tetra[0:2]
  elif typ == "222":
    return ktw_tetra[0:4]
  elif typ == "42":
    return ktw_tetra[:]
  elif typ == "4":
    return ktw_tetra[[0,3,4,5]]
  elif typ == "3":
    return ktw_hexagonal[0:3]
  elif typ == "6":
    return ktw_hexagonal[0:6]
  elif typ == "32":
    return ktw_hexagonal[[0,1,2,9,10,11]]
  elif typ == "622":
    return ktw_hexagonal[:]
  elif typ == "23":
    return ktw_cubic[0:12]
  elif typ == "432":
    return ktw_cubic[:]
  else:
    raise ValueError("Unknown rotational group %s." % typ)

groups = ["1", "2", "222", "42", "4", "3", "6", "32", "622", "23", "432"]

class TestRotationalGroups(unittest.TestCase):
  def compare(self, matrix, quaternion):
    for i,(M, q) in enumerate(zip(matrix, quaternion)):
      self.assertTrue(np.allclose(M, q.to_matrix()))

  def test_definition(self):
    for grp in groups:
      print(grp)
      self.compare(ktw_rotational_groups(grp),
          crystallography.symmetry_rotations(grp))

  def test_is_group(self):
    for grp in groups:
      print(grp)
      ops = crystallography.symmetry_rotations(grp)
      self.assertTrue(self.includes_identity(ops))
      self.assertTrue(self.closed(ops))
      self.assertTrue(self.has_inverse(ops))

  def includes_identity(self, ops):
    for op in ops:
      if np.allclose(op.quat, [1,0,0,0]):
        break
    else:
      return False
    return True

  def closed(self, ops):
    for i,op_a in enumerate(ops):
      for j,op_b in enumerate(ops):
        c = op_a * op_b
        found = False
        for op_c in ops:
          if np.allclose(c.quat, op_c.quat) or np.allclose(c.quat, -op_c.quat):
            found = True
            break
        if not found:
          return False

    return True

  def has_inverse(self, ops):
    for op_a in ops:
      opa_inv = op_a.inverse()
      for op_b in ops:
        if np.allclose(opa_inv.quat, op_b.quat) or np.allclose(opa_inv.quat, -op_b.quat):
          break
      else:
        return False
    
    return True

class LTests(object):
  def test_lattice(self):
    test = np.zeros((3,3))

    avs = [self.lattice.a1, self.lattice.a2, self.lattice.a3]
    bvs = [self.lattice.b1, self.lattice.b2, self.lattice.b3]

    for i in range(3):
      for j in range(3):
        test[i,j] = avs[i].dot(bvs[j])
    
    self.assertTrue(np.allclose(np.eye(3), test))

class ShearTests(object):
  def test_M(self):
    for i in range(self.lattice.ngroup):
      for j in range(self.lattice.nslip(i)):
        self.assertEqual(
            tensors.Symmetric(
              np.dot(self.QM,np.dot(np.outer(self.lattice.slip_directions[i][j].data,
                self.lattice.slip_planes[i][j].data), self.QM.T))),
              self.lattice.M(i,j,self.Q))

  def test_W(self):
    for i in range(self.lattice.ngroup):
      for j in range(self.lattice.nslip(i)):
        self.assertEqual(
            tensors.Skew(
              np.dot(self.QM,np.dot(np.outer(self.lattice.slip_directions[i][j].data,
                self.lattice.slip_planes[i][j].data), self.QM.T))),
              self.lattice.N(i,j,self.Q))

  def test_shear(self):
    for i in range(self.lattice.ngroup):
      for j in range(self.lattice.nslip(i)):
        self.assertTrue(np.isclose(
            np.einsum('ij,ij',
              np.dot(self.QM,np.dot(np.outer(self.lattice.slip_directions[i][j].data,
                self.lattice.slip_planes[i][j].data), self.QM.T)), self.S),
              self.lattice.shear(i,j,self.Q,self.ST)))

  def test_dshear(self):
    for i in range(self.lattice.ngroup):
      for j in range(self.lattice.nslip(i)):
        rs = lambda S: np.einsum('ij,ij',
              np.dot(self.QM,np.dot(np.outer(self.lattice.slip_directions[i][j].data,
                self.lattice.slip_planes[i][j].data), self.QM.T)), S)
        num = tensors.Symmetric(differentiate(rs, self.S)[0])

class TestCubicFCC(unittest.TestCase, LTests, ShearTests):
  def setUp(self):
    self.a = 1.3
    self.lattice = crystallography.CubicLattice(self.a)

    self.S = np.array([[20.0,-15.0,12.0],[-15.0,-40.0,5.0],[12.0,5.0,60.0]])
    self.ST = tensors.Symmetric(self.S)
    self.Q = rotations.Orientation(30.0,43.0,10.0, angle_type = "degrees")
    self.QM = self.Q.to_matrix()

    self.lattice.add_slip_system([1,1,0],[1,1,1])

  def test_fcc(self):
    self.assertEqual(len(self.lattice.burgers_vectors[0]), 12)
    self.assertEqual(len(self.lattice.slip_directions[0]), 12)
    self.assertEqual(len(self.lattice.slip_planes[0]), 12)

    self.assertEqual(self.lattice.ngroup, 1)
    self.assertEqual(self.lattice.nslip(0), 12)
    self.assertEqual(self.lattice.ntotal, 12)

    for d,n in zip(self.lattice.slip_directions[0], self.lattice.slip_planes[0]):
      self.assertAlmostEqual(d.norm(), 1.0)
      self.assertAlmostEqual(n.norm(), 1.0)
      self.assertAlmostEqual(d.dot(n), 0.0)
    
    for b in self.lattice.burgers_vectors[0]:
      self.assertAlmostEqual(b.norm(), np.sqrt(2.0) * self.a)

class TestCubicBCC(unittest.TestCase, LTests, ShearTests):
  def setUp(self):
    self.a = 1.3
    self.lattice = crystallography.CubicLattice(self.a)

    self.S = np.array([[20.0,-15.0,12.0],[-15.0,-40.0,5.0],[12.0,5.0,60.0]])
    self.ST = tensors.Symmetric(self.S)
    self.Q = rotations.Orientation(30.0,43.0,10.0, angle_type = "degrees")
    self.QM = self.Q.to_matrix()

    self.lattice.add_slip_system([1,1,1],[1,1,0])
    self.lattice.add_slip_system([1,1,1],[1,2,3])
    self.lattice.add_slip_system([1,1,1],[1,1,2])

  def test_bcc(self):
    self.assertEqual(len(self.lattice.slip_directions[0]), 12)
    self.assertEqual(len(self.lattice.slip_directions[1]), 24)
    self.assertEqual(len(self.lattice.slip_directions[2]), 12)

    self.assertEqual(self.lattice.ngroup, 3)
    self.assertEqual(self.lattice.nslip(0), 12)
    self.assertEqual(self.lattice.nslip(1), 24)
    self.assertEqual(self.lattice.nslip(2), 12)
    self.assertEqual(self.lattice.ntotal, 48)

    for dg,ng,bg in zip(self.lattice.slip_directions, self.lattice.slip_planes,
        self.lattice.burgers_vectors):
      for d,n in zip(dg, ng):
        self.assertAlmostEqual(d.norm(), 1.0)
        self.assertAlmostEqual(n.norm(), 1.0)
        self.assertAlmostEqual(d.dot(n), 0.0)

      for b in bg:
        self.assertAlmostEqual(b.norm(), np.sqrt(3.0) * self.a)





from neml import solvers, interpolate, models, elasticity, ri_flow, hardening, surfaces, parse, visco_flow, general_flow, creep, damage

import unittest
import numpy as np

from common import *

class TestErrors(unittest.TestCase):
  def test_badobject(self):
    with self.assertRaises(RuntimeError):
      test = parse.parse_xml(localize("examples.xml"), "test_badobject")

class CompareMats(object):
  def test_same(self):
    t_n = 0.0
    stress_n1 = np.zeros((6,))
    hist_n1 = self.model1.init_store()

    stress_n2 = np.zeros((6,))
    hist_n2 = self.model2.init_store()

    strain_n = np.zeros((6,))

    u_n1 = 0.0
    u_n2 = 0.0
    p_n1 = 0.0
    p_n2 = 0.0

    for i,m in enumerate(np.linspace(0,1,self.nsteps)):
      t_np1 = self.tmax * m
      strain_np1 = self.emax * m
      
      stress_np11, hist_np11, A_np11, u_np11, p_np11 = self.model1.update_sd(
          strain_np1, strain_n, self.T, self.T, t_np1, t_n, stress_n1, hist_n1,
          u_n1, p_n1)

      stress_np12, hist_np12, A_np12, u_np12, p_np12 = self.model2.update_sd(
          strain_np1, strain_n, self.T, self.T, t_np1, t_n, stress_n2, hist_n2,
          u_n2, p_n2)

      self.assertTrue(np.allclose(stress_np11, stress_np12))
      self.assertTrue(np.allclose(hist_np11, hist_np12))
      self.assertTrue(np.allclose(A_np11, A_np12))
      self.assertTrue(np.isclose(u_np11, u_np12))
      self.assertTrue(np.isclose(p_np11, p_np12))
      
      stress_n1 = stress_np11
      hist_n1 = hist_np11
      stress_n2 = stress_np12
      hist_n2 = hist_np12

      u_n1 = u_np11
      u_n2 = u_np12
      p_n1 = p_np11
      p_n2 = p_np12

      strain_n = strain_np1

class TestPowerLawDamage(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_powerdamage")

    E = 92000.0
    nu = 0.3

    s0 = 180.0
    Kp = 1000.0

    elastic = elasticity.IsotropicLinearElasticModel(E, "youngs",
        nu, "poissons")

    surface = surfaces.IsoJ2()
    hrule = hardening.LinearIsotropicHardeningRule(s0, Kp)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    bmodel = models.SmallStrainRateIndependentPlasticity(elastic, 
        flow)

    a = 2.2
    A = 2e-5
    self.model2 = damage.NEMLPowerLawDamagedModel_sd(elastic, 
        A, a, bmodel)

    self.T = 300.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.05,0,0,0,0,0])

class TestStringParser(CompareMats, unittest.TestCase):
  def setUp(self):
    string_rep = '<test_j2iso type="SmallStrainRateIndependentPlasticity"><elastic type="IsotropicLinearElasticModel"><m1>103561.64383561644</m1><m1_type>youngs</m1_type><m2>0.2945205479452055</m2><m2_type>poissons</m2_type></elastic><flow type="RateIndependentAssociativeFlow"><surface type="IsoJ2"/><hardening type="LinearIsotropicHardeningRule"><s0>100.0</s0><K>1000.0</K></hardening></flow></test_j2iso>'

    self.model1 = parse.parse_string(string_rep)

    mu = 40000.0
    K = 84000.0

    ys = 100.0
    H = 1000.0

    elastic = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")

    surface = surfaces.IsoJ2()
    hrule = hardening.LinearIsotropicHardeningRule(ys, H)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model2 = models.SmallStrainRateIndependentPlasticity(elastic, flow)

    self.T = 300.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])

class TestJ2Iso(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml("test/examples.xml", "test_j2iso")
  
    mu = 40000.0
    K = 84000.0

    ys = 100.0
    H = 1000.0

    elastic = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")

    surface = surfaces.IsoJ2()
    hrule = hardening.LinearIsotropicHardeningRule(ys, H)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model2 = models.SmallStrainRateIndependentPlasticity(elastic, flow)

    self.T = 300.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])

class TestJ2Isocomb(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_j2isocomb")
  
    E = 150000.0
    nu = 0.3

    ys = 100.0
    H = 100.0
    r = 100.0
    d = 1000.0

    elastic = elasticity.IsotropicLinearElasticModel(E, "youngs", nu,
        "poissons")

    surface = surfaces.IsoJ2()
    hrule1 = hardening.LinearIsotropicHardeningRule(ys, H)
    hrule2 = hardening.VoceIsotropicHardeningRule(0.0, r, d)
    hrule = hardening.CombinedIsotropicHardeningRule([hrule1,hrule2])
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model2 = models.SmallStrainRateIndependentPlasticity(elastic, flow)

    self.T = 300.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])

class TestCreepPlasticity(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_creep_plasticity")
  
    A = 1.85e-10
    n = 2.5

    smodel = creep.PowerLawCreep(A, n)
    cmodel = creep.J2CreepModel(smodel)

    E = 150000.0
    nu = 0.3
    sY = 200.0
    H = E / 50.0

    elastic = elasticity.IsotropicLinearElasticModel(E, "youngs", nu,
        "poissons")
    surface = surfaces.IsoJ2()
    iso = hardening.LinearIsotropicHardeningRule(sY, H)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, iso)

    pmodel = models.SmallStrainRateIndependentPlasticity(elastic, flow)
    
    self.model2 = models.SmallStrainCreepPlasticity(elastic, pmodel, cmodel)

    self.T = 300.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])

class TestJ2Combined(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_j2comb")
  
    mu = 40000.0
    K = 84000.0

    ys = 100.0
    r = 100.0
    d = 1000.0

    KH = 1000.0

    elastic = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")

    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(ys, r, d)
    kin = hardening.LinearKinematicHardeningRule(KH)
    hrule = hardening.CombinedHardeningRule(iso, kin)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model2 = models.SmallStrainRateIndependentPlasticity(elastic, flow)

    self.T = 300.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])

class TestRIChaboche(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_nonassri")
  
    mu = 40000.0
    K = 84000.0

    ys = 100.0
    r = 100.0
    d = 1000.0
    
    cs = [5.0,10.0]
    gs = [1000.0,1000.0]
    As = [0.0, 0.0]
    ns = [1.0, 1.0]

    elastic = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")

    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(ys, r, d)
    
    gammas = [hardening.ConstantGamma(g) for g in gs]
    hmodel = hardening.Chaboche(iso, cs, gammas, As, ns)
    flow = ri_flow.RateIndependentNonAssociativeHardening(surface, hmodel)

    self.model2 = models.SmallStrainRateIndependentPlasticity(elastic, flow)

    self.T = 300.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.05,0,0,0,0,0])

class TestYaguchi(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_yaguchi")
  
    mu_poly = [-0.11834615, 115.5, 48807.69]
    K_poly = [-0.256417, 250.25, 105750.0]

    shear = interpolate.PolynomialInterpolate(mu_poly)
    bulk = interpolate.PolynomialInterpolate(K_poly)
    elastic = elasticity.IsotropicLinearElasticModel(shear, "shear",
        bulk, "bulk")
    
    vmodel = visco_flow.YaguchiGr91FlowRule()

    flow = general_flow.TVPFlowRule(elastic, vmodel)

    self.model2 = models.GeneralIntegrator(elastic, flow)

    self.T = 500.0
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])

class TestRDChaboche(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_rd_chaboche")
  
    mu = 60384.61
    K = 130833.3

    elastic = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")

    r = -80.0
    d = 3.0

    Cs = [135.0e3, 61.0e3, 11.0e3]
    gs = [5.0e4, 1100.0, 1.0]
    As = [0.0, 0.0, 0.0]
    ns = [1.0, 1.0, 1.0]
    gmodels = [hardening.ConstantGamma(g) for g in gs]

    eta = 701.0

    n = 10.5
    
    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(0.0, r, d)

    hmodel = hardening.Chaboche(iso, Cs, gmodels, As, ns)

    fluidity = visco_flow.ConstantFluidity(eta)
    
    vmodel = visco_flow.ChabocheFlowRule(surface, hmodel, fluidity, n)
    flow = general_flow.TVPFlowRule(elastic, vmodel)

    self.model2 = models.GeneralIntegrator(elastic, flow)

    self.T = 550.0 + 273.15
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])

class TestPerzyna(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_perzyna")
  
    mu = 40000.0
    K = 84000.0

    elastic = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")
    
    sy = 100.0
    r = 100.0
    d = 1000.0
    KK = 1000.0

    n = 5.0
    eta = 500.0
    
    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(sy, r, d)
    kin = hardening.LinearKinematicHardeningRule(KK)

    hmodel = hardening.CombinedHardeningRule(iso, kin)

    gmodel = visco_flow.GPowerLaw(n, eta)
    
    vmodel = visco_flow.PerzynaFlowRule(surface, hmodel, gmodel)
    flow = general_flow.TVPFlowRule(elastic, vmodel)

    self.model2 = models.GeneralIntegrator(elastic, flow)

    self.T = 550.0 + 273.15
    self.tmax = 10.0
    self.nsteps = 100
    self.emax = np.array([0.1,0,0,0,0,0])


class TestPerfect(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_perfect")

    E = [-100, 100000]
    nu = 0.3

    youngs = interpolate.PolynomialInterpolate(E)
    poissons = interpolate.ConstantInterpolate(nu)
    elastic = elasticity.IsotropicLinearElasticModel(youngs, "youngs",
        poissons, "poissons")

    surface = surfaces.IsoJ2()

    Ts = [100.0, 300.0, 500.0, 700.0]
    Sys = [1000.0, 120.0, 60.0, 30.0]

    yields = interpolate.PiecewiseLinearInterpolate(Ts, Sys)

    self.model2 = models.SmallStrainPerfectPlasticity(elastic, surface, yields)

    self.T = 550.0
    self.tmax = 10.0
    self.nsteps = 50
    self.emax = np.array([0.1,0.05,0,-0.025,0,0])

  def test_alpha(self):
    self.assertTrue(np.isclose(self.model1.alpha(self.T), 0.1))

class TestPerfectCreep(CompareMats, unittest.TestCase):
  def setUp(self):
    self.model1 = parse.parse_xml(localize("examples.xml"), "test_pcreep")

    E = [-100, 100000]
    nu = 0.3

    youngs = interpolate.PolynomialInterpolate(E)
    poissons = interpolate.ConstantInterpolate(nu)
    elastic = elasticity.IsotropicLinearElasticModel(youngs, "youngs",
        poissons, "poissons")

    surface = surfaces.IsoJ2()

    Ts = [100.0, 300.0, 500.0, 700.0]
    Sys = [1000.0, 120.0, 60.0, 30.0]

    yields = interpolate.PiecewiseLinearInterpolate(Ts, Sys)

    pmodel = models.SmallStrainPerfectPlasticity(elastic, surface, yields)

    self.T = 550.0
    self.tmax = 10.0
    self.nsteps = 50
    self.emax = np.array([0.1,0.05,0,-0.025,0,0])

    A = 1.85e-10
    n = 2.5

    smodel = creep.PowerLawCreep(A, n)
    cmodel = creep.J2CreepModel(smodel)

    self.model2 = models.SmallStrainCreepPlasticity(elastic, pmodel, cmodel)



from __future__ import division

import numpy as np
import itertools
import os.path

mandel = ((0,0),(1,1),(2,2),(1,2),(0,2),(0,1))
mandel_mults = (1,1,1,np.sqrt(2),np.sqrt(2),np.sqrt(2))

skew_inds = ((1,2),(0,2),(0,1))
skew_mults = (-1.0,1.0,-1.0)

def localize(fname):
  return os.path.join(os.path.dirname(__file__), fname)

def piece_together_fourth(Dp, Wp):
  """
    Take the skew and symmetric parts of a algorithmic tangent and piece them back together
  """
  sym_id = 0.5*(np.einsum('ik,jl', np.eye(3), np.eye(3)) + np.einsum('jk,il', np.eye(3), np.eye(3)))
  skew_id = 0.5*(np.einsum('ik,jl', np.eye(3), np.eye(3)) - np.einsum('jk,il', np.eye(3), np.eye(3)))

  return np.einsum('ijkl,klmn',sym_tensor_part(Dp), sym_id) + np.einsum('ijkl,klmn', 
      skew_tensor_part(Wp), skew_id) 

def sym_tensor_part(C):
  """
     Take a Mandel stiffness in my notation and convert it back to a full tensor
  """
  Ct = np.zeros((3,3,3,3))
  for a in range(6):
    for b in range(6):
      ind_a = itertools.permutations(mandel[a], r=2)
      ind_b = itertools.permutations(mandel[b], r=2)
      ma = mandel_mults[a]
      mb = mandel_mults[b]
      indexes = tuple(ai+bi for ai, bi in itertools.product(ind_a, ind_b))
      for ind in indexes:
        Ct[ind] = C[a,b] / ma*mb

  for i in range(3):
    for j in range(3):
      for k in range(3):
        for l in range(3):
          if l < k:
            Ct[i,j,k,l] = 0.0

  return Ct

def skew_tensor_part(C):
  """
    Take a skew stiffness in my notation and convert it back to a full tensor
  """
  Ct = np.zeros((3,3,3,3))
  for a in range(6):
    for b in range(3):
      inds_a = mandel[a]
      inds_b = skew_inds[b]
      mult_a = mandel_mults[a]
      mult_b = skew_mults[b]
      for ord_a in ((0,1),(1,0)):
        for ord_b, f in zip(((0,1),(1,0)), (1,-1)):
          ind = tuple([inds_a[aa] for aa in ord_a] + [inds_b[bb] for bb in ord_b])
          Ct[ind] = C[a,b] *  mult_a*mult_b * f

  for i in range(3):
    for j in range(3):
      for k in range(3):
        for l in range(3):
          if i != j:
            Ct[i,j,k,l] /= 2.0
          if l < k: 
            Ct[i,j,k,l] = 0.0

  return Ct

def unroll_fourth(T):
  """
    Unroll a fourth order tensor into a 9x9
  """
  M = np.zeros((9,9))

  for i in range(3):
    for j in range(3):
      a = i*3+j
      for k in range(3):
        for l in range(3):
          b = k*3+l
          M[a,b] = T[i,j,k,l]

  return M

def reroll_fourth(M):
  """
    Undo unroll_fourth
  """
  T = np.zeros((3,3,3,3))
  for a in range(9):
    i = a // 3
    j = a % 3
    for b in range(9):
      k = b // 3
      l = b % 3
      T[i,j,k,l] = M[a,b]

  return T

def ms2ts(C):
  """
    Convert a Mandel notation stiffness matrix to a full stiffness tensor.
  """
  Ct = np.zeros((3,3,3,3))
  for a in range(6):
    for b in range(6):
      ind_a = itertools.permutations(mandel[a], r=2)
      ind_b = itertools.permutations(mandel[b], r=2)
      ma = mandel_mults[a]
      mb = mandel_mults[b]
      indexes = tuple(ai+bi for ai, bi in itertools.product(ind_a, ind_b))
      for ind in indexes:
        Ct[ind] = C[a,b] / (ma*mb)

  return Ct

def ts2ms(C):
  """
    Convert a stiffness tensor into a Mandel notation stiffness matrix
  """
  Cv = np.zeros((6,6))
  for i in range(6):
    for j in range(6):
      ma = mandel_mults[i]
      mb = mandel_mults[j]
      Cv[i,j] = C[mandel[i]+mandel[j]] * ma * mb

  return Cv

def ws2ts(C):
  """
    Convert a skew notation stiffness matrix to a full stiffness tensor.
  """
  Ct = np.zeros((3,3,3,3))
  for a in range(6):
    for b in range(3):
      inds_a = mandel[a]
      inds_b = skew_inds[b]
      mult_a = mandel_mults[a]
      mult_b = skew_mults[b]
      for ord_a in ((0,1),(1,0)):
        for ord_b, f in zip(((0,1),(1,0)), (1,-1)):
          ind = tuple([inds_a[aa] for aa in ord_a] + [inds_b[bb] for bb in ord_b])
          Ct[ind] = C[a,b] / (mult_a*mult_b) * f

  return Ct

def ts2ws(C):
  """
    Convert a stiffness tensor into a skew notation stiffness matrix
  """
  Cv = np.zeros((6,3))
  for i in range(6):
    for j in range(3):
      ma = mandel_mults[i]
      mb = skew_mults[j]
      Cv[i,j] = C[mandel[i]+skew_inds[j]] * ma * mb

  return Cv

def wws2ts(C):
  """
    Convert a d_skew_d_sym matrix to a tensor
  """
  Ct = np.zeros((3,3,3,3))
  for a in range(3):
    for b in range(6):
      inds_a = skew_inds[a]
      inds_b = mandel[b]
      mult_a = skew_mults[a]
      mult_b = mandel_mults[b]
      for ord_a, f in zip(((0,1),(1,0)),(1,-1)):
        for ord_b in ((0,1),(1,0)):
          ind = tuple([inds_a[aa] for aa in ord_a] + [inds_b[bb] for bb in ord_b])
          Ct[ind] = C[a,b] / (mult_a * mult_b) * f

  return Ct

def ts2wws(C):
  """
    Convert a stiffness tensor into a d_skew_d_sym matrix
  """
  Cv = np.zeros((3,6))
  for i in range(3):
    for j in range(6):
      ma = skew_mults[i]
      mb = mandel_mults[j]
      Cv[i,j] = C[skew_inds[i] + mandel[j]] * ma * mb

  return Cv

def ts2sww(C):
  """
    Convert a stiffness tensor into a d_sym_d_skew matrix
  """
  Cv = np.zeros((6,3))
  for i in range(6):
    for j in range(3):
      ma = mandel_mults[i]
      mb = skew_mults[j]
      Cv[i,j] = C[mandel[i] + skew_inds[j]] * ma * mb

  return Cv

def sym(A):
  """
    Take a symmetric matrix to the Mandel convention vector.
  """
  return np.array([A[0,0], A[1,1], A[2,2], np.sqrt(2)*A[1,2], 
    np.sqrt(2)*A[0,2], np.sqrt(2)*A[0,1]])

def usym(v):
  """
    Take a Mandel symmetric vector to the full matrix.
  """
  return np.array([
    [v[0], v[5]/np.sqrt(2), v[4]/np.sqrt(2)],
    [v[5]/np.sqrt(2), v[1], v[3]/np.sqrt(2)],
    [v[4]/np.sqrt(2), v[3]/np.sqrt(2), v[2]]
    ])

def skew(A):
  """
    Take a skew matrix to my vector convention.
  """
  return np.array([-A[1,2], A[0,2], -A[0,1]])

def uskew(v):
  """
    Take a skew vector in my convention to the matrix
  """
  return np.array([
    [0, -v[2], v[1]],
    [v[2], 0, -v[0]],
    [-v[1], v[0], 0]
    ])

def differentiate(fn, x0, eps = 1.5e-6):
  if np.isscalar(x0):
    sx = (1,)
  else:
    sx = x0.shape

  f0 = fn(x0)

  if np.isscalar(f0):
    sf = (1,)
  else:
    sf = f0.shape

  tshape = sf + sx
  rshape = sx + sf

  D = np.zeros(rshape)

  for index in np.ndindex(sx):
    if np.isscalar(x0):
      diff = np.abs(x0 * eps)
    else:
      diff = np.abs(x0[index] * eps)
    if diff < eps:
      diff = eps
    
    if np.isscalar(x0):
      xp = diff
    else:
      xp = np.zeros(sx)
      xp[index] = diff

    fp = fn(xp+x0)

    D[index] = (fp - f0) / diff
  
  Df = np.zeros(tshape)
  # Reverse
  for ind1 in np.ndindex(sx):
    for ind2 in np.ndindex(sf):
      Df[ind2+ind1] = D[ind1+ind2]
  
  if np.isscalar(x0):
    return Df[0][0]
  else:
    return Df

def make_dev(s):
  return s - np.array([1,1,1,0,0,0]) * np.sum(s[:3]) / 3.0



#!/usr/bin/env python3

from neml import history, interpolate
from neml.math import tensors, rotations
from neml.cp import crystallography, slipharden, sliprules, inelasticity

from common import differentiate
from nicediff import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonInelastic(object):
  def test_d_p_d_stress(self):
    nd = diff_symmetric_symmetric(lambda s: self.model.d_p(s, self.Q, self.H, self.L, self.T, self.fixed),
        self.S)
    d = self.model.d_d_p_d_stress(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    self.assertEqual(nd, d)

  def test_d_p_d_history(self):
    nd = diff_symmetric_history(lambda h: self.model.d_p(self.S, self.Q, h, self.L, self.T, self.fixed),
        self.H)
    d = np.array(self.model.d_d_p_d_history(self.S, self.Q, self.H, self.L, self.T, self.fixed))
    
    self.assertTrue(np.allclose(d.T.reshape(nd.shape, order = 'F'), nd))

  def test_w_p_d_stress(self):
    nd = diff_skew_symmetric(lambda s: self.model.w_p(s, self.Q, self.H, self.L, self.T, self.fixed), 
      self.S)
    d = self.model.d_w_p_d_stress(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    self.assertEqual(nd, d)

  def test_w_p_d_history(self):
    nd = diff_skew_history(lambda h: self.model.w_p(self.S, self.Q, h, self.L, self.T, self.fixed),
        self.H)
    d = np.array(self.model.d_w_p_d_history(self.S, self.Q, self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(d.T.reshape(nd.shape, order = 'F'), nd))

  def test_d_hist_rate_d_stress(self):
    nd = diff_history_symmetric(lambda s: self.model.history_rate(s, self.Q, self.H, self.L, self.T, self.fixed),
        self.S)
    d = np.array(self.model.d_history_rate_d_stress(self.S, self.Q, self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(nd.reshape(d.shape), d))

  def test_d_hist_rate_d_hist(self):
    nd = diff_history_history(lambda h: self.model.history_rate(self.S, self.Q, h, self.L, self.T, self.fixed),
        self.H)
    d = np.array(self.model.d_history_rate_d_history(self.S, self.Q, self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(nd.reshape(d.shape), d))

class TestNoInelasticity(unittest.TestCase, CommonInelastic):
  def setUp(self):
    self.model = inelasticity.NoInelasticity()

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))

    self.T = 300.0

    self.H = history.History()

    self.fixed = history.History()

  def test_d_p(self):
    self.assertEqual(tensors.Symmetric(np.zeros((3,3))),
        self.model.d_p(self.S, self.Q, self.H, self.L, self.T,self.fixed))

  def test_w_p(self):
    self.assertEqual(tensors.Skew(np.zeros((3,3))),
        self.model.w_p(self.S, self.Q, self.H, self.L, self.T,self.fixed))

  def test_hist_rate(self):
    h1 = history.History()
    h2 = self.model.history_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)

    self.assertTrue(np.allclose(np.array(h1), np.array(h2)))

class TestAsaroInelasticity(unittest.TestCase, CommonInelastic):
  def setUp(self):
    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar("strength")
    self.H.set_scalar("strength", self.strength)

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.strengthmodel = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.model = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))

    self.T = 300.0

    self.fixed = history.History()

  def test_d_p(self):
    d = tensors.Symmetric(np.zeros((3,3)))
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        d += self.slipmodel.slip(g, i, self.S, self.Q, self.H, self.L, self.T,
            self.fixed) * self.L.M(g, i, self.Q)
    
    self.assertEqual(d, 
        self.model.d_p(self.S, self.Q, self.H, self.L, self.T, self.fixed))

  def test_w_p(self):
    w = tensors.Skew(np.zeros((3,3)))
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        w += self.slipmodel.slip(g, i, self.S, self.Q, self.H, self.L, self.T, 
            self.fixed) * self.L.N(g, i, self.Q)
    
    self.assertEqual(w, 
        self.model.w_p(self.S, self.Q, self.H, self.L, self.T, self.fixed))

  def test_hist_rate(self):
    h1 = self.slipmodel.hist_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    h2 = self.model.history_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)

    self.assertTrue(np.allclose(np.array(h1), np.array(h2)))

class TestPowerLawInelasticity(unittest.TestCase, CommonInelastic):
  def setUp(self):
    self.A = 1.0e-2
    self.n = 3.1

    self.model = inelasticity.PowerLawInelasticity(self.A, self.n)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))

    self.T = 300.0

    self.H = history.History()

    self.fixed = history.History()

  def test_seq(self):
    seq1 = np.sqrt(3.0/2.0) * self.S.dev().norm()
    seq2 = np.sqrt(3.0/2.0 * self.S.dev().contract(self.S.dev()))
    self.assertTrue(np.isclose(seq1, seq2))

  def test_d_p(self):
    seq = np.sqrt(3.0/2.0) * self.S.dev().norm()

    Dp1 = self.A*seq**self.n * self.S.dev() / seq
    Dp2 = self.model.d_p(self.S, self.Q, self.H, self.L, self.T, self.fixed)

    self.assertEqual(Dp1, Dp2)

  def test_w_p(self):
    self.assertEqual(tensors.Skew(np.zeros((3,3))),
        self.model.w_p(self.S, self.Q, self.H, self.L, self.T, self.fixed))

  def test_hist_rate(self):
    h1 = history.History()
    h2 = self.model.history_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)

    self.assertTrue(np.allclose(np.array(h1), np.array(h2)))

class TestCombinedInelasticity(unittest.TestCase, CommonInelastic):
  def setUp(self):
    self.A = 1.0e-5
    self.n = 3.1

    self.model1 = inelasticity.PowerLawInelasticity(self.A, self.n)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))

    self.T = 300.0

    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar("strength")
    self.H.set_scalar("strength", self.strength)

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.strengthmodel = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.model2 = inelasticity.AsaroInelasticity(self.slipmodel)

    self.model = inelasticity.CombinedInelasticity([self.model1, self.model2])

    self.fixed = history.History()

  def test_d_p(self):
    dp1 = self.model1.d_p(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    dp2 = self.model2.d_p(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    dp = self.model.d_p(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    self.assertEqual(dp1+dp2,dp)

  def test_w_p(self):
    wp1 = self.model1.w_p(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    wp2 = self.model2.w_p(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    wp = self.model.w_p(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    self.assertEqual(wp1+wp2,wp)

  def test_hist_rate(self):
    h1 = self.model1.history_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    h2 = self.model2.history_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)
    h = self.model.history_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)

    h3 = history.History()
    h3.add_union(h1)
    h3.add_union(h2)

    self.assertTrue(np.allclose(np.array(h), np.array(h3)))

class TestComplexInelasticity(unittest.TestCase, CommonInelastic):
  def setUp(self):
    self.strength_0 = 35.0
    self.H = history.History()
    self.H.add_scalar("strength0")
    self.H.set_scalar("strength0", self.strength_0)

    self.strength_1 = 25.0
    self.H.add_scalar("strength1")
    self.H.set_scalar("strength1", self.strength_1)

    self.tau0_0 = 10.0
    self.tau_sat_0 = 50.0
    self.b_0 = 2.5

    self.tau0_1 = 5.0
    self.tau_sat_1 = 25.0
    self.b_1 = 1.0

    self.strengthmodel = slipharden.SumSlipSingleStrengthHardening(
        [
          slipharden.VoceSlipHardening(self.tau_sat_0, self.b_0, self.tau0_0),
          slipharden.VoceSlipHardening(self.tau_sat_1, self.b_1, self.tau0_1)
          ])
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.model = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))

    self.T = 300.0

    self.fixed = history.History()



import sys
sys.path.append('..')

from neml import models, block, elasticity, surfaces, hardening, ri_flow
from common import *

import unittest
import numpy as np
import numpy.random as ra

class TestBlockEvaluate(unittest.TestCase):
  def setUp(self):
    E = 150000.0
    nu = 0.3

    ys = 100.0
    H = 1000.0

    elastic = elasticity.IsotropicLinearElasticModel(E, "youngs",
        nu, "poissons")

    surface = surfaces.IsoJ2()
    hrule = hardening.LinearIsotropicHardeningRule(ys, H)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model = models.SmallStrainRateIndependentPlasticity(elastic, flow)
    
    self.nblock = 100

  def test_batch(self):
    e_np1 = np.zeros((self.nblock,3,3))
    e_np1[:,0,0] = np.linspace(0, 0.1, self.nblock)
    e_n = np.zeros((self.nblock,3,3))
    T_np1 = np.zeros((self.nblock,))
    T_n = np.zeros((self.nblock,))
    t_np1 = 1.0
    t_n = 0.0
    s_np1 = np.zeros((self.nblock,3,3))
    s_n = np.zeros((self.nblock,3,3))
    h_np1 = np.zeros((self.nblock, self.model.nstore))
    h_n = np.zeros((self.nblock, self.model.nstore))
    A_np1 = np.zeros((self.nblock,3,3,3,3))
    u_np1 = np.zeros((self.nblock,))
    u_n = np.zeros((self.nblock,))
    p_np1 = np.zeros((self.nblock,))
    p_n = np.zeros((self.nblock,))

    # Try with the block eval function
    block.block_evaluate(self.model,
        e_np1, e_n, T_np1, T_n, t_np1, t_n, s_np1, s_n, h_np1, h_n,
        A_np1, u_np1, u_n, p_np1, p_n)

    # Now compare to item-by-item evaluation
    for i in range(self.nblock):
      s, h, A, u, p = self.model.update_sd(
          sym(e_np1[i]), sym(e_n[i]), T_np1[i], T_n[i], 
          t_np1, t_n, sym(s_n[i]), h_n[i], u_n[i], 
          p_n[i])
      self.assertTrue(np.allclose(usym(s), s_np1[i]))
      self.assertTrue(np.allclose(h, h_np1[i]))
      self.assertTrue(np.allclose(ms2ts(A), A_np1[i]))
      self.assertTrue(np.isclose(u_np1[i], u))
      self.assertTrue(np.isclose(p_np1[i], p))

mandel = ((0,0),(1,1),(2,2),(1,2),(0,2),(0,1))
mandel_mults = (1,1,1,np.sqrt(2),np.sqrt(2),np.sqrt(2))

# Make the Mandel -> tensor array
def make_M2T():
  R = np.zeros((3,3,3,3,6,6))
  for a in range(6):
    for b in range(6):
      ind_a = itertools.permutations(mandel[a], r=2)
      ind_b = itertools.permutations(mandel[b], r=2)
      ma = mandel_mults[a]
      mb = mandel_mults[b]
      indexes = tuple(ai+bi for ai, bi in itertools.product(ind_a, ind_b))
      for ind in indexes:
        R[ind+(a,b)] = 1.0 / (ma*mb)

  return R

def make_usym():
  R = np.zeros((3,3,6))
  for a in range(6):
    R[mandel[a][0],mandel[a][1],a] = 1.0/ mandel_mults[a]
    R[mandel[a][1],mandel[a][0],a] = 1.0/mandel_mults[a]
  
  return R

class TestBlockConverters(unittest.TestCase):
  def setUp(self):
    self.n = 10

  def test_t2m_array(self):
    M1 = block.t2m_array()
    M2 = make_usym()

    self.assertTrue(np.allclose(M1,M2))

  def test_m2t_array(self):
    M1 = block.m2t_array()
    M2 = make_usym().transpose((2,0,1))

    self.assertTrue(np.allclose(M1,M2))

  def test_m42t4_array(self):
    M1 = block.m42t4_array()
    M2 = make_M2T().transpose(4,5,0,1,2,3)

    self.assertTrue(np.allclose(M1,M2))

  def test_t2m(self):
    T = ra.random((self.n,3,3))
    T += T.transpose(0,2,1)

    A1 = np.zeros((self.n,6))
    for i in range(self.n):
      A1[i] = sym(T[i])

    A2 = block.t2m(T)

    self.assertTrue(np.allclose(A1,A2))

  def test_m2t(self):
    M = ra.random((self.n,6))

    A1 = np.zeros((self.n,3,3))
    for i in range(self.n):
      A1[i] = usym(M[i])
    A2 = block.m2t(M)

    self.assertTrue(np.allclose(A1,A2))

  def test_m42t4(self):
    M = ra.random((self.n,6,6))

    A1 = np.zeros((self.n,3,3,3,3))
    for i in range(self.n):
      A1[i] = ms2ts(M[i])

    A2 = block.m42t4(M)

    self.assertTrue(np.allclose(A1,A2))



#!/usr/bin/env python

from neml.math import matrix

import common

import unittest
import numpy as np

class TestVector(unittest.TestCase):
  def setUp(self):
    self.a_data = [1.0,1.1,2.3]
    self.a = matrix.FlatVector(self.a_data)

  def test_definition(self):
    self.assertTrue(np.allclose(self.a_data, self.a))

  def test_size(self):
    self.assertEqual(self.a.n, len(self.a_data))

class TestBasicMatrix(unittest.TestCase):
  def setUp(self):
    self.np_mat = np.array([[2.0,3.0,1.0],[7.1,3.1,-1.0],[2.0,-1.2,8.1]])
    self.M = matrix.SquareMatrix(3, type = "dense", 
        data = list(self.np_mat.flatten())) 

  def test_definition(self):
    self.assertTrue(np.allclose(self.np_mat, self.M))

  def test_size(self):
    self.assertEqual(self.M.m, self.np_mat.shape[0])
    self.assertEqual(self.M.n, self.np_mat.shape[1])

class TestMatVec(unittest.TestCase):
  def setUp(self):
    self.a_data = [1.0,1.1,2.3]
    self.a = matrix.FlatVector(self.a_data)

    self.b_data = [-8.0,1.1,2.2,3.3]
    self.b = matrix.FlatVector(self.b_data)

    self.M_data = np.array([[2.0,3.0,1.0],[7.1,3.1,-1.0],[2.0,-1.2,8.1]])
    self.M = matrix.SquareMatrix(3, type = "dense", 
        data = list(self.M_data.flatten())) 

  def test_correct(self):
    self.assertTrue(np.allclose(self.M.dot(self.a), np.dot(self.M_data, 
      self.a_data)))

  def test_bad_size(self):
    with self.assertRaises(ValueError):
      self.M.dot(self.b)

class TestMatDefinition(unittest.TestCase):
  def test_zero(self):
    self.assertTrue(np.allclose(matrix.SquareMatrix(4), np.zeros((4,4))))

  def test_identity(self):
    self.assertTrue(np.allclose(matrix.SquareMatrix(4, type = "identity"), 
      np.eye(4)))

  def test_diagonal(self):
    self.assertTrue(np.allclose(matrix.SquareMatrix(4, type = "diagonal",
      data = [1,2,3,4]), np.diag([1,2,3,4])))

  def test_block_diagonal(self):
    self.assertTrue(np.allclose(matrix.SquareMatrix(4, type = "diagonal_blocks",
      data = [1,2], blocks = [3,1]), np.array([
        [1,0,0,0],
        [0,1,0,0],
        [0,0,1,0],
        [0,0,0,2]])))

  def test_blocks(self):
    self.assertTrue(np.allclose(matrix.SquareMatrix(4, type = "block",
      data = [1,2,3,4], blocks=[3,1]),
      np.array([
        [1,1,1,2],
        [1,1,1,2],
        [1,1,1,2],
        [3,3,3,4]])))



#!/usr/bin/env python

from neml import history
from neml.math import rotations, tensors

import unittest

import numpy as np

class TestBasicAddGet(unittest.TestCase):
  def setUp(self):
    self.hist = history.History()
    self.hist.add_scalar("a")
    self.hist.add_vector("b")
    self.hist.add_ranktwo("C")
    self.hist.add_symmetric("D")
    self.hist.add_skew("E")
    self.hist.add_orientation("q")

  def test_exists(self):
    self.assertTrue(self.hist.contains("a"))
    self.assertFalse(self.hist.contains("alsd"))

  def test_get_correct(self):
    a = self.hist.get_scalar("a")
    b = self.hist.get_vector("b")
    C = self.hist.get_ranktwo("C")
    D = self.hist.get_symmetric("D")
    E = self.hist.get_skew("E")
    q = self.hist.get_orientation("q")

  def test_get_wrong_scalar(self):
    with self.assertRaises(RuntimeError):
      a = self.hist.get_scalar("b")

  def test_get_wrong_vector(self):
    with self.assertRaises(RuntimeError):
      a = self.hist.get_vector("a")

  def test_get_wrong_ranktwo(self):
    with self.assertRaises(RuntimeError):
      a = self.hist.get_ranktwo("b")

  def test_get_wrong_symmetric(self):
    with self.assertRaises(RuntimeError):
      a = self.hist.get_symmetric("b")

  def test_get_wrong_skew(self):
    with self.assertRaises(RuntimeError):
      a = self.hist.get_skew("b")

  def test_get_wrong_orientation(self):
    with self.assertRaises(RuntimeError):
      a = self.hist.get_orientation("b")

  def test_does_not_exist(self):
    with self.assertRaises(RuntimeError):
      c = self.hist.get_scalar("g")

class TestActuallyStores(unittest.TestCase):
  def setUp(self):
    self.scalar = 2.5
    self.vector = tensors.Vector(np.array([1.0,2.0,3.0]))
    self.ranktwo = tensors.RankTwo(np.array([
      [1.0,2.0,3.0],
      [4.0,5.0,6.0],
      [7.0,8.0,9.0]]))
    self.symmetric = tensors.Symmetric(np.eye(3))
    q = np.array([[1.0,2.0,3.0],[3.0,4.0,5.0],[6.0,7.0,8.0]])
    q = 0.5 * (q - q.T)
    self.skew = tensors.Skew(q)
    self.orientation = rotations.Orientation(30.0, 60.0, 80.0, angle_type = "degrees")

    self.hist1 = history.History()
    self.hist2 = history.History(store = False)

    self.add_all(self.hist1)
    self.add_all(self.hist2)

    self.storage = np.zeros((self.hist2.size,))
    self.hist2.set_data(self.storage)

    self.set_all(self.hist1)
    self.set_all(self.hist2)

  def add_all(self, hist):
    hist.add_scalar("scalar")
    hist.add_vector("vector")
    hist.add_ranktwo("ranktwo")
    hist.add_symmetric("symmetric")
    hist.add_skew("skew")
    hist.add_orientation("orientation")

  def set_all(self, hist):
    hist.set_scalar("scalar", self.scalar)
    hist.set_vector("vector", self.vector)
    hist.set_ranktwo("ranktwo", self.ranktwo)
    hist.set_symmetric("symmetric", self.symmetric)
    hist.set_skew("skew", self.skew)
    hist.set_orientation("orientation", self.orientation)

  def test_actually_stored(self):
    self.assertTrue(np.isclose(self.scalar, self.storage[0]))

  def test_scalar_stored(self):
    self.assertTrue(np.isclose(self.scalar, self.hist1.get_scalar("scalar")))
    self.assertTrue(np.isclose(self.scalar, self.hist2.get_scalar("scalar")))

  def test_vector_stored(self):
    self.assertEqual(self.vector, self.hist1.get_vector("vector"))
    self.assertEqual(self.vector, self.hist2.get_vector("vector"))

  def test_ranktwo_stored(self):
    self.assertEqual(self.ranktwo, self.hist1.get_ranktwo("ranktwo"))
    self.assertEqual(self.ranktwo, self.hist2.get_ranktwo("ranktwo"))

  def test_symmetric_stored(self):
    self.assertEqual(self.symmetric, self.hist1.get_symmetric("symmetric"))
    self.assertEqual(self.symmetric, self.hist2.get_symmetric("symmetric"))

  def test_skew_stored(self):
    self.assertEqual(self.skew, self.hist1.get_skew("skew"))
    self.assertEqual(self.skew, self.hist2.get_skew("skew"))

  def test_orientation_stored(self):
    self.assertTrue(
        np.allclose(self.orientation.quat, self.hist1.get_orientation("orientation").quat))
    self.assertTrue(
        np.allclose(self.orientation.quat, self.hist2.get_orientation("orientation").quat))

class TestVectorMath(unittest.TestCase):
  def setUp(self):
    self.scalar = 2.5
    self.vector = tensors.Vector(np.array([1.0,2.0,3.0]))
    self.ranktwo = tensors.RankTwo(np.array([
      [1.0,2.0,3.0],
      [4.0,5.0,6.0],
      [7.0,8.0,9.0]]))
    self.symmetric = tensors.Symmetric(np.eye(3))
    q = np.array([[1.0,2.0,3.0],[3.0,4.0,5.0],[6.0,7.0,8.0]])
    q = 0.5 * (q - q.T)
    self.skew = tensors.Skew(q)

    self.hist = history.History()
    self.add_all(self.hist)
    self.set_all(self.hist)

  def add_all(self, hist):
    hist.add_scalar("scalar")
    hist.add_vector("vector")
    hist.add_ranktwo("ranktwo")
    hist.add_symmetric("symmetric")
    hist.add_skew("skew")

  def set_all(self, hist):
    hist.set_scalar("scalar", self.scalar)
    hist.set_vector("vector", self.vector)
    hist.set_ranktwo("ranktwo", self.ranktwo)
    hist.set_symmetric("symmetric", self.symmetric)
    hist.set_skew("skew", self.skew)
  
  def test_add_all(self):
    mult = 2.2
    self.hist.scalar_multiply(mult)
    self.assertTrue(np.isclose(self.hist.get_scalar("scalar"), 
      mult * self.scalar))
    self.assertEqual(self.hist.get_vector("vector"),
        mult * self.vector)
    self.assertEqual(self.hist.get_ranktwo("ranktwo"),
        mult * self.ranktwo)
    self.assertEqual(self.hist.get_symmetric("symmetric"),
        mult * self.symmetric)
    self.assertEqual(self.hist.get_skew("skew"),
        mult * self.skew)

class TestCarefulStore(unittest.TestCase):
  def setUp(self):
    self.hist = history.History(False)
    self.hist.add_scalar("a")
    self.hist.add_scalar("b")
    self.hist.add_scalar("c")
    self.data = np.array([1.0,2.0,3.0])
    self.hist.set_data(self.data)

  def test_got_from_data(self):
    self.assertEqual(self.data[0], self.hist.get_scalar("a"))

  def test_set_data(self):
    self.hist.set_scalar("a", 2.0)
    self.assertEqual(self.hist.get_scalar("a"), 2.0)
    self.assertEqual(self.data[0], 2.0)

class TestSplitStore(unittest.TestCase):
  def setUp(self):
    self.hist = history.History()
    self.hist.add_scalar("a")
    self.hist.set_scalar("a", 1.0)
    self.hist.add_scalar("b")
    self.hist.set_scalar("b", 2.0)
    self.hist.add_scalar("c")
    self.hist.set_scalar("c", 3.0)
    self.hist.add_scalar("d")
    self.hist.set_scalar("d", 4.0)

  def test_reorder(self):
    self.hist.reorder(["b", "c", "a", "d"])
    self.assertEqual(self.hist.items, ["b", "c", "a", "d"])
    self.assertTrue(np.allclose(np.array([2.0,3.0,1.0,4.0]), np.array(self.hist)))

  def test_good_split(self):
    nhist = self.hist.split(["a", "b"])
    self.assertEqual(nhist.size, 2)
    self.assertTrue(np.allclose(np.array([3.0,4.0]), np.array(nhist)))
    self.assertEqual(nhist.items, ["c", "d"])

  def test_bad_split(self):
    with self.assertRaises(RuntimeError):
      nhist = self.hist.split(["b", "c"])

  def test_subset(self):
    subset = self.hist.subset(["a", "d"])
    self.assertTrue(np.allclose(np.array([1.0,4.0]), np.array(subset)))
    self.assertEqual(subset.items, ["a", "d"])

  def test_bad_subset(self):
    with self.assertRaises(RuntimeError):
      subset = self.hist.subset(["a", "no"])

class TestSplitNoStore(unittest.TestCase):
  def setUp(self):
    self.hist = history.History(False)
    self.hist.add_scalar("a")
    self.hist.add_scalar("b")
    self.hist.add_scalar("c")
    self.data = np.array([1.0,2.0,3.0])
    self.hist.set_data(self.data)

  def test_affect(self):
    nhist = self.hist.split(["a"])
    self.assertEqual(nhist.get_scalar("b"), 2.0)
    nhist.set_scalar("b", -2.0)
    self.assertEqual(nhist.get_scalar("b"), -2.0)
    self.assertEqual(nhist.get_scalar("b"), self.hist.get_scalar("b"))
    self.assertEqual(self.data[1], -2.0)

class TestUnion(unittest.TestCase):
  def setUp(self):
    self.scalar1 = 2.0
    self.scalar2 = 3.0
    self.vector1 = tensors.Vector(np.array([1.0,2.0,3.0]))

    self.hist1 = history.History()
    self.hist1.add_scalar("a")
    self.hist1.set_scalar("a", self.scalar1)
    self.hist1.add_vector("b")
    self.hist1.set_vector("b", self.vector1) 

    self.hist2 = history.History()
    self.hist2.add_scalar("c")
    self.hist2.set_scalar("c", self.scalar2)

  def test_union(self):
    hist = self.hist2.add_union(self.hist1)
    self.assertTrue(np.isclose(hist.get_scalar("a"), self.scalar1))
    self.assertTrue(np.isclose(hist.get_scalar("c"), self.scalar2))
    self.assertEqual(hist.get_vector("b"), self.vector1)



#!/usr/bin/env python

from neml.math import rotations, tensors
from common import ms2ts

import unittest
import numpy as np
import numpy.linalg as la

def quat_mult(a, b):
  return np.concatenate((
    [a[0] * b[0] - np.dot(a[1:], b[1:])],
    a[0] * b[1:] + b[0] * a[1:] + np.cross(a[1:], b[1:])
    ))

def quat_exp(a):
  nv = la.norm(a[1:])

  nq = np.copy(a)
  a[0] = np.cos(nv)
  a[1:] *= np.sin(nv) / nv

  return a

class BasicMath(object):
  """
    Basic math operations
  """
  def test_quat(self):
    self.assertTrue(np.allclose(self.a, self.A.quat))

  def test_norm(self):
    self.assertAlmostEqual(la.norm(self.a), self.A.norm())

  def test_opposite(self):
    self.assertTrue(np.allclose(-self.a, -self.A.quat))

  def test_conjugate(self):
    aconj = np.copy(self.a)
    aconj[1:] = -aconj[1:]
    self.assertTrue(np.allclose(aconj, self.A.conj().quat))

  def test_flip(self):
    aflip = np.copy(self.a)
    aflip[0] = -aflip[0]

  def test_inverse(self):
    Ainv = self.A.inverse()
    iden = Ainv * self.A
    self.assertTrue(np.allclose(iden.quat, [1.0,0,0,0]))

  def test_quat_multiply(self):
    pv = quat_mult(self.a, self.b)
    PQ = self.A * self.B
    self.assertTrue(np.allclose(pv, PQ.quat))

  def test_quat_division(self):
    self.assertTrue(np.allclose(
      (self.A / self.B).quat,
      quat_mult(self.a, self.B.inverse().quat)))

  def test_exp(self):
    self.assertTrue(np.allclose((self.A.exp()).quat, quat_exp(self.a)))

  def test_dot(self):
    self.assertTrue(np.isclose(self.A.dot(self.B), 
      np.dot(self.A.quat, self.B.quat)))

  def test_matrix_composition(self):
    AB = self.A * self.B
    c = rotations.Quaternion(np.dot(self.A.to_product_matrix(), self.B.quat))
    self.assertTrue(np.allclose(AB.quat, c.quat))

class ScalarMath(object):
  def test_scalar_division(self):
    self.assertTrue(np.allclose((self.A / self.s).quat, self.a / self.s))

  def test_scalar_multiply(self):
    self.assertTrue(np.allclose(self.a * self.s, (self.s * self.A).quat))
    self.assertTrue(np.allclose(self.a * self.s, (self.A * self.s).quat))

class TestHash(unittest.TestCase):
  def setUp(self):
    self.Q1 = rotations.Quaternion(np.array([1.0,-2.1,3.21,4.0]))
    self.Q2 = rotations.Quaternion(np.array([5.0,-2.1,3.21,4.0]))
    self.Q3 = rotations.Quaternion(np.array([1.0000001,-2.1,3.21,4.0]))
    self.Q4 = rotations.Quaternion(np.array([4.0,-2.1,3.21,1.0]))

  def test_same(self):
    self.assertEqual(self.Q1.hash, self.Q1.hash)

  def test_not_same(self):
    self.assertNotEqual(self.Q1.hash, self.Q2.hash)
    self.assertNotEqual(self.Q1.hash, self.Q3.hash)
    self.assertNotEqual(self.Q1.hash, self.Q4.hash)

class TestBasicMathQuaternion(unittest.TestCase, BasicMath, ScalarMath):
  def setUp(self):
    self.a = np.array([2.0,-4.0,5.0,1.0])
    self.A = rotations.Quaternion(self.a)
    self.b = np.array([-1.0,3.0,2.0,0.5])
    self.B = rotations.Quaternion(self.b)

    self.s = 2.2

class TestBasicMathCOrientation(unittest.TestCase, BasicMath):
  def setUp(self):
    self.A = rotations.Orientation(31.0, 59.0, 80.0, angle_type = "degrees",
        convention = "bunge")
    self.a = np.copy(self.A.quat)
    n = np.array([1,2.0,3])
    n /= la.norm(n)
    self.B = rotations.Orientation(n, np.pi/3, angle_type = "radians")
    self.b = np.copy(self.B.quat)

class Conversion(object):
  def test_matrix(self):
    M = self.q.to_matrix()
    q2 = rotations.Orientation(M)
    self.assertTrue(np.allclose(self.q.quat, q2.quat) or 
        np.allclose(self.q.quat, -q2.quat))

  def test_axis_angle(self):
    (axis, angle) = self.q.to_axis_angle()
    q2 = rotations.Orientation(axis, angle)
    self.assertTrue(np.allclose(self.q.quat, q2.quat) or 
        np.allclose(self.q.quat, -q2.quat))

  def test_rodrigues(self):
    rv = self.q.to_rodrigues()
    q2 = rotations.Orientation(rv)
    self.assertTrue(np.allclose(self.q.quat, q2.quat) or 
        np.allclose(self.q.quat, -q2.quat))

  def test_euler_kocks(self):
    a, b, c = self.q.to_euler()
    q2 = rotations.Orientation(a, b, c)
    self.assertTrue(np.allclose(self.q.quat, q2.quat) or 
        np.allclose(self.q.quat, -q2.quat))

  def test_euler_bunge(self):
    a, b, c = self.q.to_euler(angle_type = "degrees", convention = "bunge")
    q2 = rotations.Orientation(a, b, c, angle_type = "degrees", 
        convention = "bunge")
    self.assertTrue(np.allclose(self.q.quat, q2.quat) or 
        np.allclose(self.q.quat, -q2.quat))

  def test_hopf(self):
    a, b, c = self.q.to_hopf(angle_type = "degrees")
    q2 = rotations.Orientation(a, b, c, angle_type = "degrees",
        convention = "hopf")
    self.assertTrue(np.allclose(self.q.quat, q2.quat) or 
        np.allclose(self.q.quat, -q2.quat))

  def test_hyperspherical(self):
    a1, a2, a3 = self.q.to_hyperspherical(angle_type = "degrees")
    q2 = rotations.Orientation(a1, a2, a3, angle_type = "degrees",
        convention = "hyperspherical")
    self.assertTrue(np.allclose(self.q.quat, q2.quat) or 
        np.allclose(self.q.quat, -q2.quat))

class TestEasyCaseConversion(unittest.TestCase, Conversion):
  def setUp(self):
    self.q = rotations.Orientation(30.0, 60.0, 80.0, angle_type = "degrees")

class TestSpecialCaseConversion(unittest.TestCase, Conversion):
  def setUp(self):
    self.q = rotations.Orientation(0.0, 0.0, 10.0, angle_type = "degrees")

class TestApplyThings(unittest.TestCase):
  def setUp(self):
    self.q = rotations.Orientation(30.0, 60.0, 80.0, angle_type = "degrees")
    self.Q = self.q.to_matrix()
    self.v = np.array([1.2,-2.0,3.0])
    self.Tv = tensors.Vector(self.v)

    self.A = np.array([[4.1,2.8,-1.2],[3.1,7.1,0.2],[4,2,3]])
    self.TA = tensors.RankTwo(self.A)

    self.S = np.array([[4.1,2.8,-1.2],[3.1,7.1,0.2],[4,2,3]])
    self.S = 0.5*(self.S + self.S.T)
    self.TS = tensors.Symmetric(self.S)

    self.W = np.array([[4.1,2.8,-1.2],[3.1,7.1,0.2],[4,2,3]])
    self.W = 0.5*(self.W - self.W.T)
    self.TW = tensors.Skew(self.W)

    self.R1 = np.array([[[[ 7.09627147,  9.22330744, -1.36602973],
             [-7.86118175, -1.6342633 , -5.75516189],
             [ 2.61734248,  6.40678382,  3.37981603]],
            [[ 5.65100254, -7.88797059,  7.31396665],
             [-6.35471595,  5.67698069, -8.18795178],
             [ 9.10447016,  8.91183436, -6.65254333]],
            [[ 3.20429862,  2.99308849,  4.0035241 ],
             [-4.02440197, -4.39975872, -4.33542791],
             [ 9.36746226, -2.91156335,  4.51572032]]],
           [[[-9.23675199,  8.63546962,  6.83448027],
             [ 4.35044123,  2.24508666,  9.80054664],
             [ 0.30835223, -4.05208575,  5.68966326]],
            [[ 6.40300092, -8.25998136,  5.63566553],
             [-5.02801101,  5.64005224, -7.39586166],
             [ 5.90893633,  6.02074669,  1.37112738]],
            [[-2.68485216, -4.67660156,  3.52618441],
             [-2.52484812, -0.08561168,  3.39072868],
             [ 9.11295675,  2.63102786, -4.82285415]]],
           [[[ 8.31973154,  4.76081593,  4.38377207],
             [ 6.22896742, -3.83995097,  5.37501029],
             [-0.16770967,  7.9453854 , -4.95548491]],
            [[-5.67884611, -8.44970885, -7.42037867],
             [-5.19908193, -7.87006493,  1.65949787],
             [-3.25934672,  6.27340198,  5.98643056]],
            [[-4.20166968, -2.38276224,  3.04551936],
             [ 3.68445989, -5.84357996,  3.61183543],
             [ 1.54886677,  3.3659842 ,  6.43067337]]]])
    self.TR1 = tensors.RankFour(self.R1)

    self.SS1 = np.array([
      [ 5.99159801, -2.24342348,  0.26667281, -0.95466199,  3.98931478, -0.10846981],
      [ 1.86468226, -4.32391908, -7.82738638, -7.45008989,  5.89874777, 0.45820648],
      [-5.92565398,  2.4862829 , -6.02112389,  6.75455965,  4.65183463, 9.96900579],
      [ 0.60378883, -3.72189328, -7.63388446, -5.76559403, -0.3119789 , -1.1527258 ],
      [ 4.56813135, -6.06783828, -6.18341368,  8.06169686, -9.56928844, 9.08114655],
      [-8.25516614,  6.30663846,  7.2084381 , -7.38280703, -5.96279902, 8.9935982 ]])
    self.SSS1 = ms2ts(self.SS1)
    self.TSS1 = tensors.SymSymR4(self.SS1)

  def test_apply_vector(self):
    rv = self.q.apply(self.Tv)
    self.assertAlmostEqual(rv.norm(), self.Tv.norm())
    rrv = self.q.inverse().apply(rv)
    self.assertTrue(np.allclose(rrv.data, self.v.data))

    self.assertEqual(self.q.apply(self.Tv), tensors.Vector(np.dot(self.Q, self.v)))
  
  def test_apply_ranktwo(self):
    self.assertEqual(self.q.apply(self.TA), 
        tensors.RankTwo(np.dot(self.Q, np.dot(self.A, self.Q.T))))

  def test_apply_symmetric(self):
    self.assertEqual(self.q.apply(self.TS),
        tensors.Symmetric(np.dot(self.Q, np.dot(self.S, self.Q.T))))

  def test_apply_skew(self):
    self.assertEqual(self.q.apply(self.TW),
        tensors.Skew(np.dot(self.Q, np.dot(self.W, self.Q.T))))

  @staticmethod
  def rotate_fourth(T, Q):
    return np.einsum('ip,jq,pqrs,kr,ls', Q, Q, T, Q, Q)

  def test_apply_rankfour(self):
    self.assertEqual(self.q.apply(self.TR1), 
        tensors.RankFour(TestApplyThings.rotate_fourth(self.R1, self.Q)))

  def test_apply_symsym(self):
    self.assertEqual(self.q.apply(self.TSS1),
        tensors.RankFour(TestApplyThings.rotate_fourth(self.SSS1, self.Q)
          ).to_sym())

class TestExpIntegration(unittest.TestCase):
  def setUp(self):
    self.q = rotations.Orientation(30.0, 60.0, 80.0, angle_type = "degrees")

    self.W = np.array([[4.1,2.8,-1.2],[3.1,7.1,0.2],[4,2,3]])
    self.W = 0.5*(self.W - self.W.T)
    self.TW = tensors.Skew(self.W)

  def test_w2q(self):
    self.assertEqual(self.TW, rotations.wlog(rotations.wexp(self.TW)))

  def test_q2w(self):
    self.assertTrue(np.allclose(self.q.quat,
      rotations.wexp(rotations.wlog(self.q)).quat))

  def test_correctly_integrates(self):
    # So we are rotating about x where t is the rotation in radians
    R = lambda t: np.array([[1.0,0,0],[0,np.cos(t),-np.sin(t)],[0,np.sin(t), np.cos(t)]])
    q0 = rotations.Orientation(R(0.0))

    spin = tensors.Skew(np.array([
      [0,0,0],
      [0,0,-1.0],
      [0,1.0,0]]))
    
    nsteps = 1000 # Well it's not the most accurate integration
    ang = np.pi/4.0

    qf = rotations.Orientation(R(ang))
    
    q = q0
    for i in range(nsteps):
      q = rotations.wexp(spin * ang/nsteps) * q

    self.assertTrue(np.allclose(q.quat, qf.quat))

class TestFromVectors(unittest.TestCase):
  def setUp(self):
    self.x = tensors.Vector([0.0,0.0,1.0])
    self.y = tensors.Vector([1.0,0.0,0.0])
    self.q = rotations.Orientation(self.x,self.y)

  def test_sends_correctly(self):
    x0 = tensors.Vector([1.0,0,0])
    y0 = tensors.Vector([0.0,1.0,0.0])

    xp = self.q.apply(x0)
    self.assertEqual(self.x, xp)

    yp = self.q.apply(y0)
    self.assertEqual(self.y, yp)

class TestDistance(unittest.TestCase):
  def setUp(self):
    self.q = rotations.Orientation(30.0, 60.0, 80.0, angle_type = "degrees")

  def test_same_zero(self):
    self.assertTrue(np.isclose(self.q.distance(self.q), 0))
    self.assertTrue(np.isclose(rotations.distance(self.q,self.q), 0))

class TestOrientVectors(unittest.TestCase):
  def setUp(self):
    self.v1 = tensors.Vector([1.0,4,2])
    self.v1.normalize()
    self.v2 = tensors.Vector([2,1,-6])
    self.v2.normalize()

  def test_single_orient(self):
    q = rotations.rotate_to(self.v1, self.v2)
    v3 = q.apply(self.v1)
    self.assertEqual(self.v2, v3)

  def test_family_orient(self):
    for a in np.linspace(0,2*np.pi):
      q = rotations.rotate_to_family(self.v1, self.v2, a)
      v3 = q.apply(self.v1)
      self.assertEqual(self.v2,v3)




import sys
sys.path.append('..')

from neml.math.nemlmath import *
from common import *

from neml.math import nemlmath

import unittest
import numpy as np
import numpy.linalg as la
import numpy.random as ra

class TestLDTangent(unittest.TestCase):
  def setUp(self):
    self.sym_mandel = ra.random((6,6))
    self.sym_ten = ms2ts(self.sym_mandel)
    self.sym_99 = unroll_fourth(self.sym_ten)

    self.skew_deriv_red = ra.random((6,3))
    self.skew_deriv_ten = ws2ts(self.skew_deriv_red)
    self.skew_deriv_99 = unroll_fourth(self.skew_deriv_ten)

    self.wws_deriv_red = ra.random((3,6))
    self.wws_deriv_ten = wws2ts(self.wws_deriv_red)
    self.wws_deriv_99 = unroll_fourth(self.wws_deriv_ten)

    self.S = np.array([[10.0,-20.0,15.0],[-20.0,60.0,5.0],[15.0,5.0,20.0]])
    self.S_v = sym(self.S)

  def test_piece_back(self):
    A = transform_fourth(self.sym_mandel, self.skew_deriv_red).reshape((3,3,3,3))
    B = piece_together_fourth(self.sym_mandel, self.skew_deriv_red)

    self.assertTrue(np.allclose(A, B))

  def test_mandel_convert(self):
    self.assertTrue(np.allclose(nemlmath.mandel2full(self.sym_mandel), self.sym_99))
    self.assertTrue(np.allclose(nemlmath.full2mandel(self.sym_99), self.sym_mandel))

  def test_skew_convert(self):
    self.assertTrue(np.allclose(nemlmath.skew2full(self.skew_deriv_red), self.skew_deriv_99))
    self.assertTrue(np.allclose(nemlmath.full2skew(self.skew_deriv_99), self.skew_deriv_red))

  def test_wws_convert(self):
    self.assertTrue(np.allclose(nemlmath.wws2full(self.wws_deriv_red), self.wws_deriv_99))
    self.assertTrue(np.allclose(nemlmath.full2wws(self.wws_deriv_99), self.wws_deriv_red))

  def test_identities(self):
    sym = unroll_fourth(0.5*(np.einsum('ik,jl', np.eye(3), np.eye(3)) + np.einsum('jk,il', np.eye(3), np.eye(3))))
    skew = unroll_fourth(0.5*(np.einsum('ik,jl', np.eye(3), np.eye(3)) - np.einsum('jk,il', np.eye(3), np.eye(3)))) 

    self.assertTrue(np.allclose(sym, nemlmath.idsym()))
    self.assertTrue(np.allclose(skew, nemlmath.idskew()))

  def test_outer_mat(self):
    tensor = np.einsum('in,jm', self.S, np.eye(3)) + np.einsum('im,nj', np.eye(3), self.S) - np.einsum('mn,ij', np.eye(3), self.S)
    matrix = unroll_fourth(tensor)
    
    direct = nemlmath.truesdell_tangent_outer(self.S_v)
    self.assertTrue(np.allclose(matrix, direct))

class TestLDUpdate(unittest.TestCase):
  def setUp(self):
    self.L = np.array([[0.1,-0.2,0.5],[1.1,-0.2,0.25],[0.3,-0.4,0.2]])
    self.D = 0.5 * (self.L + self.L.T)
    self.Dv = sym(self.D)
    self.W = 0.5 * (self.L - self.L.T)
    self.Wv = skew(self.W)
    
    self.S_n = np.array([[10.0,-20.0,15.0],[-20.0,60.0,5.0],[15.0,5.0,20.0]])
    self.S_n_v = sym(self.S_n)

    self.dS = np.array([[30.0,10.0,5.0],[10.0,-15.0,15.0],[5.0,15.0,-10.0]])
    self.dS_v = sym(self.dS)

  def test_setup_right(self):
    self.assertTrue(np.allclose(self.S_n, self.S_n.T))
    self.assertTrue(np.allclose(self.dS, self.dS.T))

  def test_gives_sym(self):
    """
      Make sure the update actually gives a symmetric tensor
    """
    val = self.dS - np.einsum('ik,jk', self.S_n, self.L) - np.einsum('ik,kj', self.L, self.S_n) + np.trace(self.L
        ) * self.S_n
    self.assertTrue(np.allclose(val, val.T)) 

  def test_update_theory(self):
    """
      Test to make sure my tensor math was correct on deriving the form of the update (tensor notation)
    """
    J = np.einsum('im,jn', np.eye(3), np.eye(3))*(1.0 + np.trace(self.L)) - np.einsum('im,jn', np.eye(3), self.L) - np.einsum('im,jn',
        self.L, np.eye(3))
    Jinv = la.tensorinv(J)
    interm = self.dS + np.dot(self.S_n, self.L.T) + np.dot(self.L, self.S_n) - np.trace(self.L) * self.S_n
    S_np1 = self.S_n + np.einsum('ijkl,kl', Jinv, interm)
    
    dS_c = S_np1 - self.S_n

    verify = dS_c - np.dot(S_np1, self.L.T) - np.dot(self.L, S_np1) + np.trace(self.L) * S_np1
    self.assertTrue(np.allclose(self.dS, verify))

  def test_sym_to_vector(self):
    """
      Make sure the C++ tensor -> Mandel works
    """
    self.assertTrue(np.allclose(self.Dv, nemlmath.sym(self.D)))

  def test_sym_to_tensor(self):
    """
      Make sure the C++ Mandel -> tensor works
    """
    self.assertTrue(np.allclose(self.D, nemlmath.usym(self.Dv)))

  def test_skew_to_vector(self):
    """
      Test C++ tensor -> skew
    """
    self.assertTrue(np.allclose(self.Wv, nemlmath.skew(self.W)))

  def test_skew_to_tensor(self):
    """
      Test C++ vector -> tensor
    """
    self.assertTrue(np.allclose(self.W, nemlmath.uskew(self.Wv)))

  def test_rhs(self):
    """
      Make sure the Truesdell rhs is correct
    """
    interm = sym(self.dS + np.dot(self.S_n, self.L.T) + np.dot(self.L, self.S_n) - np.trace(self.L) * self.S_n)
    outterm = truesdell_rhs(self.Dv, self.Wv, self.S_n_v, self.dS_v)
    self.assertTrue(np.allclose(interm, outterm))

  def test_matrix(self):
    """
      Make sure the Truesdell matrix is correct
    """
    J = np.einsum('im,jn', np.eye(3), np.eye(3))*(1.0 + np.trace(self.L)) - np.einsum('im,jn', np.eye(3), self.L) - np.einsum('im,jn',
        self.L, np.eye(3))
    Jm = unroll_fourth(J)
    
    j2 = truesdell_mat(self.Dv, self.Wv)

    self.assertTrue(np.allclose(Jm, j2))

  def test_implemented_update(self):
    """
      Test to make sure the vector update function will work
    """
    J = np.einsum('im,jn', np.eye(3), np.eye(3))*(1.0 + np.trace(self.L)) - np.einsum('im,jn', np.eye(3), self.L) - np.einsum('im,jn',
        self.L, np.eye(3))
    Jinv = la.tensorinv(J)
    interm = self.dS + np.dot(self.S_n, self.L.T) + np.dot(self.L, self.S_n) - np.trace(self.L) * self.S_n
    S_np1 = self.S_n + np.einsum('ijkl,kl', Jinv, interm)
    
    S_np1_v = sym(S_np1)

    S_np1_2 = truesdell_update_sym(self.Dv, self.Wv, self.S_n_v, self.dS_v)

    self.assertTrue(np.allclose(S_np1_v, S_np1_2))

class TestAddSubtractNegate(unittest.TestCase):
  def setUp(self):
    self.n = 10
    self.a = ra.random((self.n,))
    self.b = ra.random((self.n,))

  def test_minus(self):
    acopy = np.copy(self.a)
    self.assertTrue(np.allclose(minus_vec(self.a), -acopy))

  def test_add(self):
    self.assertTrue(np.allclose(add_vec(self.a, self.b), self.a + self.b))

  def test_sub(self):
    self.assertTrue(np.allclose(sub_vec(self.a, self.b), self.a - self.b))

class TestDotNorm(unittest.TestCase):
  def setUp(self):
    self.n = 10
    self.a = ra.random((self.n,))
    self.b = ra.random((self.n,))

  def test_dot(self):
    self.assertTrue(np.isclose(dot_vec(self.a,self.b), np.dot(self.a,self.b)))

  def test_norm(self):
    self.assertTrue(np.isclose(norm2_vec(self.a), la.norm(self.a)))

  def test_normalize(self):
    nv = self.a / la.norm(self.a)
    self.assertTrue(np.allclose(nv, normalize_vec(self.a)))

class TestDev(unittest.TestCase):
  def setUp(self):
    self.s = np.array([10.0,-15.0,5.0,4.0,8.0,-12.0])

  def test_dev(self):
    sdev = self.s - np.array([1,1,1,0,0,0]) * np.sum(self.s[:3]) / 3.0
    self.assertTrue(np.allclose(sdev, dev_vec(self.s)))

class TestOuter(unittest.TestCase):
  def setUp(self):
    self.na = 10
    self.nb = 12
    self.a = ra.random((self.na,))
    self.b = ra.random((self.nb,))
    self.C = ra.random((self.na,self.nb))

  def test_outer(self):
    self.assertTrue(
        np.allclose(outer_vec(self.a, self.b), np.outer(self.a, self.b)),
        msg = str(outer_vec(self.a, self.b)))

  def test_update(self):
    
    Cu = self.C + np.outer(self.a, self.b)
    Cn = outer_update(self.a, self.b, self.C)
    self.assertTrue(np.allclose(Cn, Cu))

  def test_update_minus(self):
    Cu = self.C - np.outer(self.a, self.b)
    self.assertTrue(
        np.allclose(outer_update_minus(self.a, self.b, self.C), Cu))

class TestMatVec(unittest.TestCase):
  def setUp(self):
    self.m = 10
    self.n = 8
    self.A = ra.random((self.m, self.n))
    self.b = ra.random((self.n,))
    self.c = ra.random((self.m,))

  def test_matvec(self):
    self.assertTrue(np.allclose(np.dot(self.A, self.b), 
      mat_vec(self.A, self.b)),
      msg = str(np.dot(self.A, self.b)) + '\n' + str(mat_vec(self.A, self.b)))

  def test_matvec_trans(self):
    self.assertTrue(np.allclose(np.dot(self.A.T, self.c),
      mat_vec_trans(self.A, self.c)))

class TestMatMat(unittest.TestCase):
  def setUp(self):
    self.m = 10
    self.n = 8
    self.k = 9
    self.A = ra.random((self.m, self.k))
    self.B = ra.random((self.k, self.n))

  def test_matmat(self):
    self.assertTrue(np.allclose(np.dot(self.A, self.B), mat_mat(self.A, self.B)))

  def test_matmat_ABT(self):
    B = self.B.T
    self.assertTrue(np.allclose(np.dot(self.A, B.T), mat_mat_ABT(self.A, B)))

class TestRotateMat(unittest.TestCase):
  def setUp(self):
    self.m = 6
    self.n = 6
    self.A = ra.random((self.m,self.n))
    self.B = ra.random((self.n,self.n))

  def test_rotate(self):
    C = nemlmath.rotate_matrix(self.A, self.B)
    C2 = np.dot(self.A, np.dot(self.B, self.A.T))
    self.assertTrue(np.allclose(C,C2))

class TestInvert(unittest.TestCase):
  def setUp(self):
    self.n = 10
    self.square_ns = ra.random((self.n,self.n))
    self.nonsquare = ra.random((self.n,self.n-1))
    self.big = ra.random((self.n,self.n,self.n))

  def test_invert(self):
    inv_ns = la.inv(self.square_ns)
    inv = invert_mat(self.square_ns)
    self.assertTrue(np.allclose(inv_ns, inv))

  def test_nonsquare(self):
    self.assertRaises(RuntimeError, invert_mat, self.nonsquare)

  def test_nonmatrix(self):
    self.assertRaises(RuntimeError, invert_mat, self.big)

class TestSolve(unittest.TestCase):
  def setUp(self):
    self.n = 10
    self.A = ra.random((self.n,self.n))
    self.b = ra.random((self.n,))

  def test_solve(self):
    x = la.solve(self.A, self.b)
    self.b = solve_mat(self.A, self.b)
    print(x)
    print(self.b)
    self.assertTrue(np.allclose(x, self.b))

class TestDiagSolve(unittest.TestCase):
  def setUp(self):
    self.n = 10
    self.dl = ra.random((self.n-1,))
    self.d = ra.random((self.n,))
    self.du = ra.random((self.n-1))
    self.A = np.zeros((self.n, self.n))
    for i in range(self.n):
      self.A[i,i] = self.d[i]
      if i != self.n -1:
        self.A[i,i+1] = self.du[i]
        self.A[i+1,i] = self.dl[i]
    
    self.c = ra.random((self.n,))

  def test_solve(self):
    x_np = la.solve(self.A, self.c)
    dl, d, du, du2, ipiv = dgttrf(self.dl, self.d, self.du)
    x_la = dgttrs(dl, d, du, du2, ipiv, self.c)
    self.assertTrue(np.allclose(x_np, x_la))

class TestCond(unittest.TestCase):
  def setUp(self):
    self.n = 10
    self.A = ra.random((self.n,self.n))

  def test_cond(self):
    fr = condition(self.A)

class TestPoly(unittest.TestCase):
  def setUp(self):
    self.n = 10
    self.poly = ra.random((self.n,))
    self.x = ra.random((1,))[0]
  
  def test_eval(self):
    npv = np.polyval(self.poly, self.x)
    mv  = polyval(self.poly, self.x)
    self.assertTrue(np.isclose(npv, mv))

  def test_derivs(self):
    for n in range(5):
      p1 = np.polyder(self.poly, m = n)
      p2 = differentiate_poly(self.poly, n)
      self.assertTrue(np.allclose(p1,p2))

  def test_problem_case(self):
    p = np.array([-1.,  0.,  1.])
    for n in range(4):
      p1 = np.polyder(p, m = n)
      p2 = differentiate_poly(p, n)
      self.assertTrue(np.allclose(p1,p2))

  def test_poly_from_roots(self):
    p1 = np.polynomial.polynomial.polyfromroots(self.poly)[::-1]
    p2 = poly_from_roots(self.poly)

    self.assertTrue(np.allclose(p1,p2))

class TestMisc(unittest.TestCase):
  def setUp(self):
    self.a1_deg = 127.0
    self.a1_rad = np.deg2rad(self.a1_deg)

  def test_cast_angle(self):
    self.assertAlmostEqual(
        self.a1_deg, 
        nemlmath.cast_angle(self.a1_rad, "degrees"))
    self.assertAlmostEqual(
        self.a1_rad,
        nemlmath.cast_angle(self.a1_rad, "radians"))

  def test_convert_angle(self):
    self.assertAlmostEqual(
        self.a1_rad,
        nemlmath.convert_angle(self.a1_deg, "degrees"))
    self.assertAlmostEqual(
        self.a1_rad,
        nemlmath.convert_angle(self.a1_rad, "radians"))

  def test_gcd(self):
    self.assertEqual(nemlmath.gcd(12,20), 4)

  def test_common_gcd(self):
    self.assertEqual(nemlmath.common_gcd([10,15,20]), 5)

  def test_reduce_gcd(self):
    self.assertEqual(nemlmath.reduce_gcd([10,20,15]), [2,4,3])

  def test_isclose(self):
    self.assertTrue(nemlmath.isclose(1.0e10, 1.00001e10))
    self.assertFalse(nemlmath.isclose(1.0e-7, 1.0e-8))

class TestVariousFactorial(unittest.TestCase):
  def test_factorial(self):
    for i in range(10):
      self.assertTrue(np.isclose(np.math.factorial(i), factorial(i)))

class TestEigenstuff(unittest.TestCase):
  def setUp(self):
    self.S = np.array([50.0,-25.0,100.0,30.0,-180.0,90.0])

  def test_eigenvalues_sym(self):
    vals = eigenvalues_sym(self.S)
    tvs = la.eigvalsh(usym(self.S))
    
    self.assertTrue(np.allclose(np.array(vals), np.array(tvs)))

  def test_eigenvectors_sym(self):
    vecs = eigenvectors_sym(self.S)
    tvals, tvecs = la.eigh(usym(self.S))
    tvecs = tvecs.T
 
    for nv, mv in zip(tvecs,vecs):
      self.assertTrue(np.allclose(nv,mv) or np.allclose(nv,-mv))

class TestInvariants(unittest.TestCase):
  def setUp(self):
    self.S = np.array([50.0,-25.0,100.0,30.0,-180.0,90.0])
    self.SF = usym(self.S)
    self.SF_dev = self.SF - np.trace(self.SF)/3.0 * np.eye(3)
  
  def test_I1(self):
    self.assertTrue(np.isclose(np.trace(self.SF), I1(self.S)))

  def test_I2(self):
    v1 = 0.5*(np.trace(self.SF)**2.0 - np.trace(np.dot(self.SF, self.SF)))
    self.assertTrue(np.isclose(v1, I2(self.S)))

  def test_I1_dev(self):
    self.assertTrue(np.isclose(np.trace(self.SF_dev), I1(dev_vec(self.S))))

  def test_I2_dev(self):
    v1 = 0.5*(np.trace(self.SF_dev)**2.0 - np.trace(np.dot(self.SF_dev, self.SF_dev)))
    self.assertTrue(np.isclose(v1, I2(dev_vec(self.S))))



import sys
sys.path.append('..')

from neml import interpolate, solvers, models, elasticity, ri_flow, hardening, surfaces, visco_flow, general_flow, creep, uniaxial
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonUniaxial(object):
  """
    Common uniaxial model tests
  """
  def test_tangent(self):
    hn = self.umodel.init_store()
    en = 0.0
    sn = 0.0
    un = 0.0
    pn = 0.0
    tn = 0.0
    Tn = self.T0

    n = 100
    emax = 0.1

    for i in range(self.nsteps):
      enp1 = en + self.de
      Tnp1 = Tn + self.dT
      tnp1 = tn + self.dt
      snp1, hnp1, Anp1, unp1, pnp1 = self.umodel.update(enp1, en, Tnp1, Tn, 
          tnp1, tn, sn, hn, un, pn)
      
      sfn = lambda e: self.umodel.update(e, en, Tnp1, Tn, tnp1, tn,
          sn, hn, un, pn)[0]
      nA = differentiate(sfn, enp1)
      
      self.assertTrue(np.isclose(nA, Anp1, rtol = 1.0e-3))

      sn = snp1
      hn = np.copy(hnp1)
      en = enp1
      Tn = Tnp1
      tn = tnp1
      un = unp1
      pn = pnp1

class TestUniaxialRI(CommonUniaxial, unittest.TestCase):
  """
    Test with a simple uniaxial model
  """
  def setUp(self):
    E = 200000.0
    nu = 0.27

    mu = E / (2 * (1.0 + nu))
    K = E / (3 * (1 - 2 * nu))

    s0 = 300.0
    Kp = 0.0
    c = [30000.0]
    r = [60.0]
    A = [0.0]
    n = [1.0]
    
    elastic = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")
    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(s0, Kp)
    gmodels = [hardening.ConstantGamma(g) for g in r]
    hrule = hardening.Chaboche(iso, c, gmodels, A, n)

    flow = ri_flow.RateIndependentNonAssociativeHardening(surface, hrule)
    self.model = models.SmallStrainRateIndependentPlasticity(elastic, flow, verbose = False)
    self.umodel = uniaxial.UniaxialModel(self.model)
    
    self.de = 0.001
    self.dt = 1.0
    self.dT = 0.0
    self.T0 = 0.0
    self.nsteps = 100



import sys
sys.path.append('..')

from neml.surfaces import *
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonYieldSurface(object):
  """
    Inherit from me and define setUp to run standard tests over a yield
    surface.
  """
  def test_dfds(self):
    dfn = lambda s: self.model.f(s, self.hist, self.T)
    n_d = differentiate(dfn, self.s)
    d = self.model.df_ds(self.s, self.hist, self.T)
    self.assertTrue(np.allclose(n_d, d))

  def test_dfdq(self):
    dfn = lambda q: self.model.f(self.s, q, self.T)
    n_d = differentiate(dfn, self.hist)
    self.assertTrue(np.allclose(n_d, 
      self.model.df_dq(self.s, self.hist, self.T)))

  def test_dfdsds(self):
    dfn = lambda s: self.model.df_ds(s, self.hist, self.T)
    n_d = differentiate(dfn, self.s)
    self.assertTrue(np.allclose(n_d,
      self.model.df_dsds(self.s, self.hist, self.T)))
    
  def test_dfdqdq(self):
    dfn = lambda q: self.model.df_dq(self.s, q, self.T)
    n_d = differentiate(dfn, self.hist)
    d = self.model.df_dqdq(self.s, self.hist, self.T)
    self.assertTrue(np.allclose(n_d, d))

  def test_dfdsdq(self):
    dfn = lambda q: self.model.df_ds(self.s, q, self.T)
    n_d = differentiate(dfn, self.hist)
    self.assertTrue(np.allclose(n_d,
      self.model.df_dsdq(self.s, self.hist, self.T)))

  def test_dfdqds(self):
    dfn = lambda s: self.model.df_dq(s, self.hist, self.T)
    n_d = differentiate(dfn, self.s)
    self.assertTrue(np.allclose(n_d,
      self.model.df_dqds(self.s, self.hist, self.T)))

class TestIsoJ2(unittest.TestCase, CommonYieldSurface):
  def setUp(self):
    self.s = np.array([10.0,-50.0,250.0,25.0,33.0,-40.0])
    self.hist = np.array([-150.0])
    self.T = 300.0

    self.model = IsoJ2()

  def test_hist(self):
    self.assertEqual(self.model.nhist, 1)

  def test_f(self):
    sdev = self.s - np.array([1,1,1,0,0,0]) * np.sum(self.s[:3])/3
    exact = la.norm(sdev) + np.sqrt(2.0/3.0) * self.hist[0]
    self.assertTrue(np.isclose(self.model.f(self.s, self.hist, self.T),
      exact), 
      msg = str(self.model.f(self.s, self.hist, self.T)))

class TestIsoKinJ2(unittest.TestCase, CommonYieldSurface):
  def setUp(self):
    self.s = np.array([100.0,-500.0,250.0,25.0,33.0,-40.0])
    self.hist = np.array([-150.0,-50.0,10.0,100.0,-10.0,30.0,-50.0])
    self.hist[1:] -= np.array([1,1,1,0,0,0]) * np.sum(self.hist[1:4]) / 3.0
    self.T = 300.0

    self.model = IsoKinJ2()

  def test_hist(self):
    self.assertEqual(self.model.nhist, 7)

  def test_f(self):
    sdev = self.s - np.array([1,1,1,0,0,0]) * np.sum(self.s[:3])/3
    sdev += self.hist[1:]
    exact = la.norm(sdev) + np.sqrt(2.0/3.0) * self.hist[0]
    self.assertTrue(np.isclose(self.model.f(self.s, self.hist, self.T),
      exact))

class TestIsoKinJ2I1(unittest.TestCase, CommonYieldSurface):
  def setUp(self):
    self.s = np.array([100.0,-500.0,250.0,25.0,33.0,-40.0])
    self.hist = np.array([-150.0,-50.0,10.0,100.0,-10.0,30.0,-50.0])
    self.hist[1:] -= np.array([1,1,1,0,0,0]) * np.sum(self.hist[1:4]) / 3.0
    self.T = 300.0

    self.h = 5.0e-3
    self.l = 2.0

    self.model = IsoKinJ2I1(self.h, self.l)

  def test_hist(self):
    self.assertEqual(self.model.nhist, 7)

  def test_f(self):
    sdev = self.s - np.array([1,1,1,0,0,0]) * np.sum(self.s[:3])/3
    I1 = np.sum(self.s[:3])
    sdev += self.hist[1:]
    exact = la.norm(sdev) + np.sqrt(2.0/3.0) * self.hist[0] + self.h * np.abs(I1) ** self.l * np.sign(I1)
    self.assertTrue(np.isclose(self.model.f(self.s, self.hist, self.T),
      exact))

class TestIsoJ2I1(unittest.TestCase, CommonYieldSurface):
  def setUp(self):
    self.s = np.array([100.0,-500.0,250.0,25.0,33.0,-40.0])
    self.hist = np.array([-150.0])
    self.T = 300.0

    self.h = 5.0e-3
    self.l = 2.0

    self.model = IsoJ2I1(self.h, self.l)

  def test_hist(self):
    self.assertEqual(self.model.nhist, 1)
  
  def test_f(self):
    sdev = self.s - np.array([1,1,1,0,0,0]) * np.sum(self.s[:3])/3
    I1 = np.sum(self.s[:3])
    exact = la.norm(sdev) + np.sqrt(2.0/3.0) * self.hist[0] + self.h * np.abs(I1) ** self.l * np.sign(I1)
    self.assertTrue(np.isclose(self.model.f(self.s, self.hist, self.T),
      exact))



import sys
sys.path.append('..')

from neml import ri_flow, surfaces, hardening
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonFlowRule(object):
  """
    Tests common to all flow rules.
  """
  def gen_stress(self):
    s = np.array([50.0,75.0,-50.0,25.0,75.0,-25.0])
    return s

  def test_history(self):
    self.assertEqual(self.model.nhist, len(self.hist0))
    self.assertTrue(np.allclose(self.model.init_hist(), self.hist0))

  def test_df_ds(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.f(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.df_ds(stress, hist, self.T)
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_df_da(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.f(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.df_da(stress, hist, self.T)
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_ds(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.g(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dg_ds(stress, hist, self.T)
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_da(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.g(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dg_da(stress, hist, self.T)
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_ds(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.h(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dh_ds(stress, hist, self.T)
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_da(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.h(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dh_da(stress, hist, self.T)
    
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))


class TestRateIndependentAssociativeFlowJ2Linear(unittest.TestCase, CommonFlowRule):
  def setUp(self):
    self.s0 = 200.0
    self.K = 15000.0

    self.surface = surfaces.IsoJ2()
    self.hardening = hardening.LinearIsotropicHardeningRule(self.s0, self.K)
    self.model = ri_flow.RateIndependentAssociativeFlow(self.surface, self.hardening)

    self.hist0 = np.zeros((1,))
    
    self.T = 300.0

  def gen_hist(self):
    return 0.25
  
  def test_flow_rule(self):
    stress = self.gen_stress()
    dev_stress = make_dev(stress)
    should_be = dev_stress / la.norm(dev_stress)
    
    self.assertTrue(np.allclose(should_be, 
      self.model.g(stress, self.gen_hist(), self.T)))

  def test_hardening_rule(self):
    should_be = np.array([np.sqrt(2.0/3.0)])
    self.assertTrue(np.allclose(should_be,
      self.model.h(self.gen_stress(), self.gen_hist(), self.T)))


class TestRIChabocheLinear(unittest.TestCase, CommonFlowRule):
  """
    Test the rate independent Chaboche model with linear isotropic hardening
  """
  def setUp(self):
    self.s0 = 200.0
    self.K = 15000.0

    self.n = 4
    self.cs = range(1,self.n+1)
    self.rs = range(1, self.n+1)

    self.iso = hardening.LinearIsotropicHardeningRule(self.s0, self.K)
    self.gmodels = [hardening.ConstantGamma(g) for g in self.rs]
    As = [0.0] * self.n
    ns = [1.0] * self.n
    self.hardening = hardening.Chaboche(self.iso, self.cs, self.gmodels, As, 
        ns)
    self.surface = surfaces.IsoKinJ2()

    self.model = ri_flow.RateIndependentNonAssociativeHardening(self.surface, self.hardening)

    self.hist0 = np.zeros((1+6*self.n,))

    self.T = 300.0

  def gen_hist(self):
    hist = np.array(range(1,2+self.n*6))  / (self.n*7)
    hist[1:] = (1.0 - 2.0 * hist[1:]) * 100.0
    for i in range(self.n):
      hist[1+i*6:1+(i+1)*6] = make_dev(hist[1+i*6:1+(i+1)*6])
    return hist

  def test_flow_rule(self):
    stress = self.gen_stress()
    s = make_dev(stress)
    hist = self.gen_hist()

    X = sum(hist[1+i*6:1+(i+1)*6] for i in range(self.n))

    n = (s+X) / la.norm(s+X)

    self.assertTrue(np.allclose(n,
      self.model.g(stress, hist, self.T)))



#!/usr/bin/env python3

import argparse, time, pickle
import os.path
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))

from neml import models, parse, drivers

xml = os.path.join(os.path.dirname(__file__),"reference.xml")
models = os.path.join(os.path.dirname(__file__),"models.txt")

strain_rate = 1.0e-4

align = 50

def run_model(model_name, xml_file, repeat):
  model = parse.parse_xml(xml_file, model_name)
  
  fstr = "    Running %s..." % model_name
  print(fstr, end = '')

  start = time.time()
  for i in range(repeat):
    res = drivers.uniaxial_test(model, strain_rate)
  end = time.time()

  avg = (end-start) / repeat
  
  pad = " " * (align - len(fstr))
  print(pad+"%f s (avg)" % avg)

  return res

if __name__ == "__main__":
  parser = argparse.ArgumentParser(description="Run regression tests.")
  parser.add_argument('--regenerate', action = 'store_true', 
      help = "Regenerate the reference data")
  parser.add_argument('--xml-file',
      default = xml, help = "XML file with the models")
  parser.add_argument('--model-file',
      default = models, help = "Text file with the list to run")
  parser.add_argument('--single-model',
      help = "If set only run a single model given by the argument")
  parser.add_argument('--repeat', default = 1, type = int, 
      help = "Repeat each test several times for timing information")

  args = parser.parse_args()

  run = [line.strip() for line in open(args.model_file)]
  
  if args.single_model:
    if args.single_model in run:
      run = [args.single_model]
    else:
      raise ValueError("Requested single model %s is not in the list"
          "of valid models!" % args.single_model)
  
  print("Running models:")
  results = {model: run_model(model, args.xml_file, args.repeat)
      for model in run}

  if args.regenerate:
    for k, v in results.items():
      fname = k + ".pickle"
      pickle.dump(v, open(fname, 'wb'))





import sys
sys.path.append('..')

import unittest

import numpy as np

from neml import solvers

class TestConvergence(unittest.TestCase):
  def setUp(self):
    self.A = 1.0
    self.n = 4.0
    self.b = -16.0
    self.x0 = 4.0
    self.model = solvers.TestPower(self.A, self.n, self.b, self.x0)
    self.ts = solvers.TrialState()

  def test_correct(self):
    x = solvers.solve(self.model, self.ts, rtol = 1.0e-20, atol = 1.0e-12,
        linesearch = False, verbose = False)
    self.assertAlmostEqual(x[0], (-self.b/self.A)**(1.0/self.x0))

  def test_rtol(self):
    R0, J0 = self.model.RJ(np.array([self.x0]), self.ts)

    x = solvers.solve(self.model, self.ts, rtol = 1.0e-4, atol = 1.0e-40,
        linesearch = False, verbose = False)
    Rf, Rj = self.model.RJ(x, self.ts)
    self.assertTrue(Rf[0]/R0[0] < 1.0e-4)

  def test_atol(self):
    R0, J0 = self.model.RJ(np.array([self.x0]), self.ts)

    x = solvers.solve(self.model, self.ts, rtol = 1.0e-40, atol = 1.0e-4,
        linesearch = False, verbose = False)
    Rf, Rj = self.model.RJ(x, self.ts)
    self.assertTrue(Rf[0] < 1.0e-4)



import sys
sys.path.append('..')

from neml import elasticity, interpolate
from neml.math import tensors, rotations
import unittest

from common import *

import numpy as np
import numpy.linalg as la

class CommonElasticity(object):
  """
    Tests that could apply to any elastic model.
  """
  def test_C2S(self):
    C = self.model.C(self.T)
    self.assertTrue(np.allclose(self.model.S(self.T), la.inv(C)))

  def test_S2C(self):
    S = self.model.S(self.T)
    self.assertTrue(np.allclose(self.model.C(self.T), la.inv(S)))

  def test_tensor_C(self):
    CT = self.model.C_tensor(self.T)
    C = self.model.C(self.T)
    self.assertEqual(CT, tensors.SymSymR4(C))

  def test_tensor_S(self):
    ST = self.model.S_tensor(self.T)
    S = self.model.S(self.T)
    self.assertEqual(ST, tensors.SymSymR4(S))

class TestIsotropicConstantModel(CommonElasticity, unittest.TestCase):
  def setUp(self):
    self.mu = 29000.0
    self.K = 64000.0
    self.T = 325.0

    self.Q = rotations.Orientation(31.0, 59.0, 80.0, angle_type = "degrees",
        convention = "bunge")

    self.model = elasticity.IsotropicLinearElasticModel(self.mu, 
        "shear", self.K, "bulk")

  def test_modulii(self):
    S = self.model.S(self.T)
    E = 9*self.K*self.mu/(3*self.K + self.mu)
    nu = (3*self.K - 2*self.mu)/(2*(3*self.K+self.mu))

    self.assertTrue(np.isclose(S[0,0], 1/E))
    self.assertTrue(np.isclose(S[0,1], -nu/E))
    self.assertTrue(np.isclose(S[3,3], (1+nu)/E))

  def test_rotated(self):
    no = self.model.C_tensor(self.T)
    yes = self.model.C_tensor(self.T, self.Q)

    self.assertEqual(no,yes)

class TestEquivalentDefinitions(unittest.TestCase):
  def setUp(self):
    self.E = 100000.0
    self.nu = 0.3

    self.mu = self.E / (2 * (1 + self.nu))
    self.K = self.E / (3 * (1 - 2.0 * self.nu))
    
    self.model_Ev = elasticity.IsotropicLinearElasticModel(
        self.E, "youngs", self.nu, "poissons")
    self.model_GK = elasticity.IsotropicLinearElasticModel(
        self.mu, "shear", self.K, "bulk")

    self.T = 300.0

  def test_equivalent_C(self):
    C1 = self.model_Ev.C(self.T)
    C2 = self.model_GK.C(self.T)
    self.assertTrue(np.allclose(C1,C2))

  def test_equivalent_S(self):
    S1 = self.model_Ev.S(self.T)
    S2 = self.model_GK.S(self.T)
    self.assertTrue(np.allclose(S1,S2))
  
  def test_equivalent_modulii(self):
    self.assertTrue(np.isclose(self.E, self.model_Ev.E(self.T)))
    self.assertTrue(np.isclose(self.E, self.model_GK.E(self.T)))
    self.assertTrue(np.isclose(self.nu, self.model_Ev.nu(self.T)))
    self.assertTrue(np.isclose(self.nu, self.model_GK.nu(self.T)))
    self.assertTrue(np.isclose(self.mu, self.model_Ev.G(self.T)))
    self.assertTrue(np.isclose(self.mu, self.model_GK.G(self.T)))
    self.assertTrue(np.isclose(self.K, self.model_Ev.K(self.T)))
    self.assertTrue(np.isclose(self.K, self.model_GK.K(self.T)))

class TestCubicModel(CommonElasticity, unittest.TestCase):
  def setUp(self):
    self.mu = 29000.0
    self.E = 120000.0
    self.nu = 0.3
    self.T = 325.0

    self.Q = rotations.Orientation(31.0, 59.0, 80.0, angle_type = "degrees",
        convention = "bunge")
    self.Q_cube = rotations.Orientation(90.0, 0.0, 0.0, angle_type = "degrees",
        convention = "bunge") 

    self.model = elasticity.CubicLinearElasticModel(self.E, 
        self.nu, self.mu, "moduli")

    self.v1 = tensors.Vector(np.array([1.0,0.0,0]))
    self.v2 = tensors.Vector(np.array([0.0,1.0,0]))

  def test_definition(self):
    self.assertTrue(np.isclose(self.model.G(self.T), self.mu))
    self.assertTrue(np.isclose(self.model.G(self.T,
      self.Q_cube, self.v1, self.v2), self.mu))

    self.assertEqual(self.model.C_tensor(self.T),
        self.model.C_tensor(self.T, self.Q_cube))



import sys
sys.path.append('..')

from common import *

from neml import interpolate
import unittest

import numpy as np
import numpy.random as ra
import scipy.interpolate as inter

class BaseInterpolate(object):
  def test_derivative(self):
    d = self.interpolate.derivative(self.x)
    nd = differentiate(lambda x: self.interpolate(x), self.x)
    self.assertTrue(np.isclose(d, nd, rtol = 1.0e-3))

class TestPolynomialInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.n = 5
    self.coefs = list((ra.random((self.n,)) * 0.5 - 1.0) * 2.0)
    self.x = ra.random((1,))[0]
    self.interpolate = interpolate.PolynomialInterpolate(self.coefs)

  def test_interpolate(self):
    self.assertTrue(np.isclose(np.polyval(self.coefs, self.x),
      self.interpolate(self.x)))

class TestPiecewiseLinearInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.validx = [-10.0, -2.0, 1.0, 2.0, 5.0, 15.0]
    self.invalidx_1 = [-2.0, -10.0, 1.0, 2.0, 5.0, 15.0]
    self.invalidx_2 = [-10.0, -2.0, 1.0, 2.0, 5.0]
    self.points = [50.0, -25.0, 1.0, 0.0, 0.0, 10.0]

    self.x = 17.0

    self.valid = interpolate.PiecewiseLinearInterpolate(self.validx, 
        self.points)
    self.invalid1 = interpolate.PiecewiseLinearInterpolate(self.invalidx_1, 
        self.points)
    self.invalid2 = interpolate.PiecewiseLinearInterpolate(self.invalidx_2, 
        self.points)
    self.interpolate = self.valid

  def test_valid(self):
    self.assertTrue(self.valid.valid)
    self.assertFalse(self.invalid1.valid)
    self.assertFalse(self.invalid2.valid)

  def test_interpolate(self):
    testinter = inter.interp1d(self.validx, self.points,
        bounds_error = False)
    xs = np.linspace(-15.0,20.0)
    ys1 = [self.valid(x) for x in xs]
    ys2 = testinter(xs)
    ys2[xs < self.validx[0]] = self.points[0]
    ys2[xs > self.validx[-1]] = self.points[-1]
    self.assertTrue(np.allclose(ys1, ys2))

class TestGenericPiecewiseInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.xs = [1.0,5.0]
    self.cvs = [-1.0,5.0,-2.0]
    self.fns = [interpolate.ConstantInterpolate(cv) for cv in self.cvs]

    self.interpolate = interpolate.GenericPiecewiseInterpolate(self.xs,
        self.fns)

    self.x = 4.0

  def test_interpolate(self):
    self.assertTrue(np.isclose(self.interpolate.value(0.0), self.cvs[0]))
    self.assertTrue(np.isclose(self.interpolate.value(4.0), self.cvs[1]))
    self.assertTrue(np.isclose(self.interpolate.value(10.0), self.cvs[2]))

class TestPiecewiseLogLinearInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.validx = [-10.0, -2.0, 1.0, 2.0, 5.0, 15.0]
    self.invalidx_1 = [-2.0, -10.0, 1.0, 2.0, 5.0, 15.0]
    self.invalidx_2 = [-10.0, -2.0, 1.0, 2.0, 5.0]
    self.points = [1.0e5, 1.0e0, 1.0e-1, 1.0e-6, 1.0e0, 1.0e2]
    self.invalidpoints = [1.0e5, 1.0e0, -1.0e-1, 1.0e-6, 1.0e0, 1.0e2]

    self.x = 17.0

    self.valid = interpolate.PiecewiseLogLinearInterpolate(self.validx, 
        self.points)
    self.invalid1 = interpolate.PiecewiseLogLinearInterpolate(self.invalidx_1, 
        self.points)
    self.invalid2 = interpolate.PiecewiseLogLinearInterpolate(self.invalidx_2, 
        self.points)
    self.invalid3 = interpolate.PiecewiseLogLinearInterpolate(self.validx, 
        self.invalidpoints)
    self.interpolate = self.valid

  def test_valid(self):
    self.assertTrue(self.valid.valid)
    self.assertFalse(self.invalid1.valid)
    self.assertFalse(self.invalid2.valid)
    self.assertFalse(self.invalid3.valid)

  def test_interpolate(self):
    testinter = inter.interp1d(self.validx, np.log(self.points),
        bounds_error = False)
    xs = np.linspace(-15.0,20.0)
    ys1 = [self.valid(x) for x in xs]
    yp = testinter(xs)
    yp[xs < self.validx[0]] = np.log(self.points[0])
    yp[xs > self.validx[-1]] = np.log(self.points[-1])
    ys2 = np.exp(yp)
    self.assertTrue(np.allclose(ys1, ys2))

class TestPiecewiseSemiLogXLinearInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.validx = [1e-2, 0.5, 1.0, 2.0, 5.0, 15.0]
    self.points = [10,20,30,40,50,60]

    self.x = 7.5

    self.interpolate = interpolate.PiecewiseSemiLogXLinearInterpolate(self.validx, 
        self.points)

  def test_interpolate(self):
    testinter = inter.interp1d(np.log10(self.validx), self.points,
        bounds_error = False)
    xs = np.linspace(1e-3,20.0,500)
    ys1 = [self.interpolate(x) for x in xs]
    ys2 = testinter(np.log10(xs))
    ys2[xs < self.validx[0]] = self.points[0]
    ys2[xs > self.validx[-1]] = self.points[-1]
    self.assertTrue(np.allclose(ys1, ys2))

class TestConstantInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.v = ra.random((1,))[0]
    self.interpolate = interpolate.ConstantInterpolate(self.v)
    self.x = ra.random((1,))[0]

  def test_interpolate(self):
    self.assertTrue(np.isclose(self.v, self.interpolate(self.x)))

class TestExpInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.A = 1.2
    self.B = 5.1
    self.interpolate = interpolate.ExpInterpolate(self.A, self.B)
    self.x = 0.5

  def test_interpolate(self):
    self.assertTrue(np.isclose(self.A * np.exp(self.B/self.x),
      self.interpolate(self.x)))

class TestPowerLawInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.A = 5.1
    self.B = 0.3
    self.interpolate = interpolate.PowerLawInterpolate(self.A, self.B)
    self.x = 2.1

  def test_interpolate(self):
    self.assertTrue(np.isclose(self.A * self.x**self.B, 
      self.interpolate(self.x)))

class TestMTSInterpolate(unittest.TestCase, BaseInterpolate):
  def setUp(self):
    self.y0 = 100.0
    self.D = 50.0
    self.x0 = 200.0

    self.interpolate = interpolate.MTSShearInterpolate(self.y0, self.D, self.x0)

    self.x = 300.0

  def test_interpolate(self):
    should = self.y0 - self.D / (np.exp(self.x0 / self.x) - 1.0)
    act = self.interpolate(self.x)
    self.assertTrue(np.isclose(should, act))
